import pika
import json
import pprint
import sys

def main():
    import argparse
    usage = "ank_measure_client"
    parser = argparse.ArgumentParser(description = usage)
    parser.add_argument('--hostname',  default= "measure_client", help="Hostname for messaging")   
    parser.add_argument('--server', '-s',  default= None, help="RabbitMQ server")   
    arguments = parser.parse_args()

    server = arguments.hostname
    try:
        pika_host = arguments.server # test if manually specified
    except IndexError:
        import autonetkit.config as config
        pika_host = config.settings['Rabbitmq']['server']

    connection = pika.BlockingConnection(pika.ConnectionParameters(
            host = pika_host))
    channel = connection.channel()

    channel.exchange_declare(exchange='measure',
            type='direct')

    result = channel.queue_declare(exclusive=True)
    queue_name = result.method.queue

    channel.queue_bind(exchange='measure',
                       queue=queue_name,
                       routing_key = server)

    print ' [*] Waiting for messages. To exit press CTRL+C'

    def callback(ch, method, properties, body):
        data = json.loads(body)
        command = data['command']
        hosts = data['hosts']
        threads = data['threads']
        run_command(channel, command, hosts,  threads)

    channel.basic_consume(callback,
                      queue=queue_name,
                      no_ack=True)

    channel.start_consuming()

def run_command(rmq_channel, command, hosts,  threads):
    """Execute command on remote host"""
    from Exscript import Account, Host
    from Exscript.util.start import start

    results = {}

    def do_something(thread, host, conn):
        #conn.execute('vtysh -c "%s"' % command)
        conn.execute(command)
        result = repr(conn.response)
        data = { host.address: {command: result}, }
        body = json.dumps(data)
        rmq_channel.basic_publish(exchange='measure',
                routing_key = "result",
                body= body)
        #conn.execute("exit")
        
    accounts = [Account("root", password = "1234")] 

    hosts = [Host(h, default_protocol = "ssh") for h in hosts]
#TODO: open multiple ssh sessions
    start(accounts, hosts, do_something, max_threads = threads)
    return results

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



import autonetkit.log as log
import pika
import json
import autonetkit.plugins.process_data as process_data
import autonetkit.config as config
import autonetkit.ank_messaging as ank_messaging
import autonetkit.log as log

def send(nidb, command, hosts, server = "measure_client", threads = 3):
# netaddr IP addresses not JSON serializable
    hosts = [str(h) for h in hosts]
    log.debug("Measure: %s %s" % (hosts, command))

    pika_host = config.settings['Rabbitmq']['server']

    messaging = ank_messaging.AnkMessaging()

    try:
        connection = pika.BlockingConnection(pika.ConnectionParameters(
            host= pika_host))
        channel = connection.channel()

        channel.exchange_declare(exchange='measure',
                type='direct')
    except pika.exceptions.AMQPConnectionError:
        log.warning("Unable to connect to RabbitMQ on %s, exiting measurement" % pika_host)
        return

    data = {
            'command': command,
            "hosts": hosts,
            "threads": threads,
            }

    body = json.dumps(data)
    channel.basic_publish(exchange='measure',
            routing_key = server,
            body= body)
    #connection.close()

    hosts_received = set(hosts)

    # parsing function mappings
    parsing = {
            'vtysh -c "show ip route"': process_data.sh_ip_route,
            "traceroute": process_data.traceroute,
            }
    
    parse_result = []

    # wait for responses
    result = channel.queue_declare(exclusive=True)
    queue_name = result.method.queue
    channel.queue_bind(exchange='measure',
                       queue=queue_name,
                       routing_key="result")

    for method_frame, properties, body in channel.consume(queue_name):
        data = json.loads(body)
        completed = False
        for host, host_data in data.items():
            for command, command_result in host_data.items():
                command_result = command_result.replace("\\r\\n", "\n")
                if command in parsing:
                    log.info( "%s %s" % (host, command))
                    parse_command = parsing[command]
                    host = process_data.reverse_tap_lookup(nidb, host)
                    parse_result = parse_command(host, nidb, command_result)
                    completed = True

                elif "traceroute" in command:
                    dst = command.split()[-1]   # last argument is the dst ip
                    src_host = process_data.reverse_tap_lookup(nidb, host)
                    dst_host = process_data.reverse_lookup(nidb, dst)
                    log.info("Trace from %s to %s" % (src_host, dst_host[1]))
                    parse_command = parsing["traceroute"]
                    log.info(command_result)
                    trace_result = parse_command(src_host, nidb, command_result)
                    trace_result.insert(0, src_host) 
                    log.info(trace_result)
                    parse_result.append(trace_result)
                    if str(trace_result[-1]) == str(dst_host[1]): #TODO: fix so direct comparison, not string, either here or in anm object comparison: eg compare on label?
#TODO: make this use custom ANK serializer function
                        trace_result = [str(t.id) for t in trace_result if t] # make serializable
                        import autonetkit.ank_messaging
                        autonetkit.ank_messaging.highlight([], [], paths = [trace_result])
                    else:
                        log.info("Partial trace, not sending to webserver: %s", trace_result)
                elif "show ip ospf interface" in command:
                    print command_result
                    completed = True
                elif "conf t" in command:
                    print command_result
                    completed = True
                else:
                    print "No parser defined for command %s" % command
                    print "Raw output:"
                    print command_result
                    completed = True

        if host in hosts_received:
            hosts_received.remove(host) # remove from list of waiting hosts

        if not len(hosts_received):
            completed = True

        if completed:
            break

    return parse_result





from autonetkit import console_script

#console_script.main()

import autonetkit
autonetkit.update_http()

import autonetkit.load.graphml as graphml
import autonetkit.anm

anm =  autonetkit.ANM()
#input_graph = graphml.load_graphml("small_internet.graphml")
input_graph = graphml.load_graphml("example/singleas.graphml")
g_in = anm.initialise_graph(input_graph)
autonetkit.update_http(anm)



# based on http://reminiscential.wordpress.com/2012/04/07/realtime-notification-delivery-using-rabbitmq-tornado-and-websocket/
import tornado
import tornado.websocket as websocket
import os
import json
import glob
import sys
import autonetkit.config as ank_config
import logging
import pkg_resources
import socket
www_dir = pkg_resources.resource_filename(__name__, "www_vis")

class MyWebHandler(tornado.web.RequestHandler):

    def initialize(self, ank_accessor):
        self.ank_accessor = ank_accessor
    
    def get(self):
        self.write("Hello, world")

    def post(self):
        data = self.get_argument('body', 'No data received')
        data_type = self.get_argument('type', 'No data received')
        print "Received data of type %s" % data_type
        #self.write(data)

        if data_type == "anm":
            body_parsed = json.loads(data)
            if False: # use to save the default.json
                import gzip
                with gzip.open(os.path.join(www_dir, "default.json.gz"), "w") as fh:
                    json.dump(body_parsed, fh)
            self.anm = data
            try:
                self.ank_accessor.anm = body_parsed
                #TODO: could process diff and only update client if data has changed -> more efficient client side
                self.update_listeners("overlays")
                # TODO: find better way to replace object not just local reference, as need to replace for RequestHandler too
            except Exception, e:
                print "Exception is", e

        elif data_type == "ip_allocations":
            body_parsed = json.loads(data)
            alloc = body_parsed
            self.ank_accessor.ip_allocation = alloc
            self.update_listeners("ip_allocations")

        elif data_type == "starting_host":
            self.update_listeners(data) 
            for listener in self.application.socket_listeners:
                #TODO: use a json format of {'type': type, 'data': data} in client-side script
                listener.write_message({'starting': data}) 

        elif data_type == "lab started":
            self.update_listeners(data) 
            for listener in self.application.socket_listeners:
                listener.write_message({'lab started': data}) 

        elif data_type == "highlight":
            #self.update_listeners(data) 
            body_parsed = json.loads(data)
            for listener in self.application.socket_listeners:
                #listener.write_message(data) 
                listener.write_message({'highlight': body_parsed}) 
        else:
            self.update_listeners(data)

        #for listener in self.application.socket_listeners:
            #listener.write_message(data) 

    def update_listeners(self, index):
        for listener in self.application.socket_listeners:
            if index == "overlays":
                listener.update_overlay()
            elif index == "ip_allocations":
                listener.update_ip_allocation()

    
class OverlayHandler(tornado.web.RequestHandler):
    def initialize(self, ank_accessor):
        self.ank_accessor = ank_accessor

    def get(self):
        overlay_id = self.get_argument("id")
        if overlay_id == "*":
            overlay_list = sorted(self.ank_accessor.overlays())
            self.write(json.dumps({'overlay_list': overlay_list}))
            return

class MyWebSocketHandler(websocket.WebSocketHandler):
    def initialize(self, ank_accessor, overlay_id):
        """ Store the overlay_id this listener is currently viewing.
        Used when updating."""
        self.ank_accessor = ank_accessor
        self.overlay_id = overlay_id

    def allow_draft76(self):
        # for iOS 5.0 Safari
        return True

    def open(self, *args, **kwargs):
        print "Client connected from %s" % self.request.remote_ip
        self.application.socket_listeners.add(self) 

        try:
            self.application.pc.add_event_listener(self)
        except AttributeError:
            pass # no RabbitMQ server

    def on_close(self):
        self.application.socket_listeners.remove(self) 
        print "Client disconnected from %s" % self.request.remote_ip
        try:
            self.application.pc.remove_event_listener(self)
        except AttributeError:
            pass # no RabbitMQ server
        try:
            self.application.echo_server.remove_event_listener(self)
        except AttributeError:
            pass # no echo_server

    def on_message(self, message):
        if "overlay_id" in message:
            _, overlay_id = message.split("=") #TODO: form JSON on client side, use loads here
            self.overlay_id = overlay_id
            self.update_overlay()
        elif "overlay_list" in message:
            body = json.dumps({'overlay_list': self.ank_accessor.overlays()})
            self.write_message(body)
        elif "ip_allocations" in message:
            body = json.dumps({'ip_allocations': self.ank_accessor.ip_allocations()})
            self.write_message(body)

    def update_overlay(self):
        body = self.ank_accessor[self.overlay_id]
        self.write_message(body)
# and update overlay dropdown
        body = json.dumps({'overlay_list': self.ank_accessor.overlays()})
        self.write_message(body)
#TODO: tidy up the passing of IP allocations

    def update_ip_allocation(self):
        body = json.dumps({'ip_allocations': self.ank_accessor.ip_allocations()})
        self.write_message(body)

class AnkAccessor():
    """ Used to store published topologies"""
    def __init__(self):
        self.anm = {}
        self.ip_allocation = {}
# try loading from vis directory
        try:
            import gzip
            fh = gzip.open(os.path.join(www_dir, "default.json.gz"), "r")
            data = json.load(fh)
            #data = json.loads(loaded)
            self.anm = data
        except IOError, e:
            print e
            pass # use default blank anm

    def overlays(self):
        if not len(self.anm):
            return [""]
        return sorted(self.anm.keys())

    def __getitem__(self, key):
        try:
            return self.anm[key]
        except KeyError:
            return json.dumps(["No ANM loaded"])

    def ip_allocations(self):
        return self.ip_allocation


 
def main():
    ank_accessor = AnkAccessor()
# check if most recent outdates current most recent

    content_path = www_dir # default content directory

    try:
        import autonetkit_cisco
    except ImportError:
        pass  # use AutoNetkit internal web content
    else:
        # use web content from autonetkit_cisco module
        content_path = pkg_resources.resource_filename("autonetkit_cisco", "web_content")

    print content_path
    settings = {
            "static_path": content_path,
            'debug': False,
            "static_url_prefix": "test",
            }

    application = tornado.web.Application([
        (r'/ws', MyWebSocketHandler, {"ank_accessor": ank_accessor, "overlay_id": "phy"}),
        (r'/publish', MyWebHandler, {"ank_accessor": ank_accessor}),
        (r'/overlay', OverlayHandler, {'ank_accessor': ank_accessor}),
        ("/(.*)", tornado.web.StaticFileHandler, {"path":settings['static_path'], "default_filename":"index.html"} )
        ], **settings)


    application.socket_listeners = set() # TODO: see if tornado provides access to listeners

    io_loop = tornado.ioloop.IOLoop.instance()
    # PikaClient is our rabbitmq consumer
  
    #listening for web clientshost_address
#TODO: make this driven from config
    try:
        port = sys.argv[1]
    except IndexError:
        port = 8000
    try:
        application.listen(port)
    except socket.error, e:
        if e.errno is 48: # socket in use
            logging.warning("Unable to start webserver: socket in use for port %s" % port)
            raise SystemExit


    print "Visualisation server started"
    io_loop.start()

if __name__ == '__main__':
    main()




import networkx as nx
import itertools
import pprint
import time
from autonetkit.ank_utils import unwrap_edges, unwrap_nodes
import autonetkit.log as log
import functools
import string

try:
    import cPickle as pickle
except ImportError:
    import pickle

class AutoNetkitException(Exception):
    pass

class OverlayNotFound(AutoNetkitException):
    def __init__(self, errors):
        self.Errors = errors

    def __str__(self):
        return "Overlay %s not found" % self.Errors

#TODO: rename to OverlayInterface
class overlay_interface(object):
    def __init__(self, anm, overlay_id, node_id, interface_id):
        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
        object.__setattr__(self, 'node_id', node_id)
        object.__setattr__(self, 'interface_id', interface_id)

    def __repr__(self):
        description = self.description or self.interface_id
        return "(%s, %s)" % (self.node_id, description)

    def __eq__(self, other):
        return (self.node_id, self.interface_id) == (other.node_id, other.interface_id)

    def __nonzero__(self):
        #TODO: work out why description and type being set/copied to each overlay
        try:
            interface = self._interface
        except KeyError:
            return False

        return len(interface) > 0  # if interface data set

    def __lt__(self, other):
        #TODO: check how is comparing the nodes
        return ((self.node, self.interface_id) < (other.node, other.interface_id))

    @property
    def is_bound(self):
        """Returns if this interface is bound to an edge on this layer"""
        return len(self.edges()) > 0

    def __str__(self):
        return self.__repr__()

    @property
    def _graph(self):
        """Return graph the node belongs to"""
        return self.anm.overlay_nx_graphs[self.overlay_id]

    @property
    def _node(self):
        """Return graph data the node belongs to"""
        return self._graph.node[self.node_id]

    @property
    def _interface(self):
        """Return data dict for the interface"""
        return self._node["_interfaces"][self.interface_id]

    @property
    def phy(self):
        # check overlay requested exists
        if self.overlay_id == "phy":
            return self
        return overlay_interface(self.anm, 'phy', 
                self.node_id, self.interface_id)

    def __getitem__(self, overlay_id):
        """Returns corresponding interface in specified overlay"""
        if not self.anm.has_overlay(overlay_id):
            log.warning("Trying to access interface %s for non-existent overlay %s"
                    % (self, overlay_id))
            return None

        if not self.node_id in self.anm.overlay_nx_graphs[overlay_id]:
            log.debug("Trying to access interface %s for non-existent node %s in overlay %s" 
                    % (self, self.node_id, self.overlay_id))
            return None

        try:
            return overlay_interface(self.anm, overlay_id, self.node_id, self.interface_id)
        except KeyError:
            return

    @property
    def is_loopback(self):
        """"""
        return self.type == "loopback" or self.phy.type == "loopback"

    @property
    def is_physical(self):
        """"""
        return self.type == "physical" or self.phy.type == "physical"

    @property
    def description(self):
        """"""
        retval = self._interface.get("description")
        if retval:
            return retval

        if self.overlay_id != "phy":  # prevent recursion
            self.phy._interface.get("description")

    @property
    def is_loopback_zero(self):
        return self.interface_id == 0 and self.is_loopback

    @property
    def type(self):
        """"""
#TODO: make 0 correctly access interface 0 -> copying problem
# TODO: this needs a bugfix rather than the below hard-coded workaround
        if self.interface_id == 0:
            return "loopback"

        if self.overlay_id != "phy":  # prevent recursion
            return self.phy._interface.get("type")

        retval = self._interface.get("type")
        if retval:
            return retval

        if self.overlay_id != "phy":  # prevent recursion
            return self.phy._interface.get("type")

    @property
    def node(self):
        """Returns parent node of this interface"""
        return OverlayNode(self.anm, self.overlay_id, self.node_id)

    def dump(self):
        return str(self._interface.items())

    def __getattr__(self, key):
        """Returns interface property"""
        try:
            return self._interface.get(key)
        except KeyError:
            return

    def get(self, key):
        """For consistency, node.get(key) is neater 
        than getattr(interface, key)"""
        return getattr(self, key)

    def __setattr__(self, key, val):
        """Sets interface property"""
        try:
            self._interface[key] = val
        except KeyError:
            self.set(key, val)

    def set(self, key, val):
        """For consistency, node.set(key, value) is neater
        than setattr(interface, key, value)"""
        return self.__setattr__(key, val)

    def edges(self):
        """Returns all edges from node that have this interface ID
        This is the convention for binding an edge to an interface"""
        # edges have _interfaces stored as a dict of {node_id: interface_id, }
        valid_edges = [e for e in self.node.edges()
                if self.node_id in e._interfaces
                and e._interfaces[self.node_id] == self.interface_id]
        return valid_edges

@functools.total_ordering
class OverlayNode(object):
    """OverlayNode"""

    def __init__(self, anm, overlay_id, node_id):
# Set using this method to bypass __setattr__
        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
# should be able to use _graph from here as anm and overlay_id are defined
        object.__setattr__(self, 'node_id', node_id)

    def __hash__(self):
        """"""
        return hash(self.node_id)

    def __nonzero__(self):
        """Allows for checking if node exists"""
        return self.node_id in self._graph

    def __iter__(self):
        """Shortcut to iterate over the physical interfaces of this node"""
        return self.interfaces(type="physical")

    def __getnewargs__(self):
        """"""
        return ()

    def __getstate__(self):
        """For pickling"""
        return (self.anm, self.overlay_id, self.node_id)

    def __setstate__(self, state):
        """"""
        # Make self.history = state and last_change and value undefined
        (anm, overlay_id, node_id) = state
        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
        object.__setattr__(self, 'node_id', node_id)

    def __eq__(self, other):
        """"""
        try:
            return self.node_id == other.node_id
        except AttributeError:
            return self.node_id == other # eg compare Node to label

    @property
    def loopback_zero(self):
        """"""
        return (i for i in self.interfaces("is_loopback_zero")).next()

    @property
    def physical_interfaces(self):
        """"""
        return self.interfaces(type = "physical")

    @property
    def loopback_interfaces(self):
        """"""
        return self.interfaces(type = "loopback")

    def __lt__(self, other):
        """"""
# want [r1, r2, ..., r11, r12, ..., r21, r22] not [r1, r11, r12, r2, r21, r22]
# so need to look at numeric part
        self_node_id = self.node_id
        other_node_id = other.node_id
        self_node_string = [x for x in self.node_id if x not in string.digits]
        other_node_string = [x for x in self.node_id if x not in string.digits]
        if self_node_string == other_node_string:
            self_node_id = "".join(
                [x for x in self.node_id if x in string.digits])
            other_node_id = "".join(
                [x for x in other.node_id if x in string.digits])
            try:
                self_node_id = int(self_node_id)
            except ValueError:
                pass  # not a number
            try:
                other_node_id = int(other_node_id)
            except ValueError:
                pass  # not a number

        return ((self.asn, self_node_id) < (other.asn, other_node_id))

    @property
    def _next_int_id(self):
        """"""
# returns next free interface ID
        for int_id in itertools.count(1):  # start at 1 as 0 is loopback
            if int_id not in self._interfaces:
                return int_id

    # TODO: interface function access needs to be cleaned up
    def _add_interface(self, description=None, type="physical", *args, **kwargs):
        """"""
        data = dict(kwargs)

        if self.overlay_id != 'phy' and self.phy:
            next_id = self.phy._next_int_id
            self.phy._interfaces[next_id] = {'type': type,
                                             'description': description} 
            #TODO: fix this workaround for not returning description from phy graph
            data['description'] = description
        else:
            next_id = self._next_int_id
            data['type'] = type  # store type on node
            data['description'] = description

        self._interfaces[next_id] = data
        return next_id

    def add_loopback(self, *args, **kwargs):
        """Public function to add a loopback interface"""
        interface_id = self._add_interface(type="loopback", *args, **kwargs)
        return overlay_interface(self.anm, self.overlay_id, 
                self.node_id, interface_id)

    def add_interface(self,*args,  **kwargs):
        """Public function to add interface"""
        interface_id = self._add_interface(*args, **kwargs)
        return overlay_interface(self.anm, self.overlay_id, 
                self.node_id, interface_id)

    def interfaces(self, *args, **kwargs):
        """Public function to view interfaces"""
        def filter_func(interface):
            """Filter based on args and kwargs"""
            return (
                all(getattr(interface, key) for key in args) and
                all(getattr(
                    interface, key) == val for key, val in kwargs.items())
            )

        all_interfaces = iter(overlay_interface(self.anm, self.overlay_id,
            self.node_id, interface_id)
            for interface_id in self._interface_ids)

        retval = (i for i in all_interfaces if filter_func(i))
        return retval

    def interface(self, key):
        """Returns interface based on interface id"""
        try:
            if key.interface_id in self._interface_ids:
                return overlay_interface(self.anm, self.overlay_id, 
                        self.node_id, key.interface_id)
        except AttributeError:
            log.warning("Unable to find interface %s in %s " % (key, self))
            return None

    @property
    def _interface_ids(self):
        """Returns interface ids for this node"""
        if self.overlay_id != 'phy' and self.phy:
            # graph isn't physical, and node exists in physical graph -> use
            # the interface mappings from phy
            return self.phy._graph.node[self.node_id]["_interfaces"].keys()
        else:
            try:
                return self._graph.node[self.node_id]["_interfaces"]
            except KeyError:
                log.debug("No interfaces initialised for %s" % self)
                return []

    @property
    def _interfaces(self):
        """Returns underlying interface dict"""
        try:
            return self._graph.node[self.node_id]["_interfaces"]
        except KeyError:
            log.debug("No interfaces initialised for %s" % self)
            return []

    @property
    def _graph(self):
        """Return graph the node belongs to"""
        return self.anm.overlay_nx_graphs[self.overlay_id]

    @property
    def is_router(self):
        """Either from this graph or the physical graph"""
        return self.device_type == "router" or self.phy.device_type == "router"

    @property
    def is_switch(self):
        """Returns if device is a switch"""
        return self.device_type == "switch" or self.phy.device_type == "switch"

    @property
    def is_server(self):
        """Returns if device is a server"""
        return self.device_type == "server" or self.phy.device_type == "server"

    @property
    def is_l3device(self):
        """Layer 3 devices: router, server, cloud, host
        ie not switch
        """
        return self.is_router or self.is_server

    def __getitem__(self, key):
        """Get item key"""
        return OverlayNode(self.anm, key, self.node_id)

    @property
    def asn(self):
        """Returns ASN of this node"""
        try:
            return self._graph.node[self.node_id]['asn']  # not in this graph
        except KeyError:
            # try from phy
            return self.anm.overlay_nx_graphs['phy'].node[self.node_id]['asn'] 

    @property
    def id(self):
        """Returns node id"""
        return self.node_id

    @property
    def _overlay(self):
        """Access overlay graph for this node"""
        return OverlayGraph(self.anm, self.overlay_id)

    def degree(self):
        """Returns degree of node"""
        return self._graph.degree(self.node_id)

    def neighbors(self, *args, **kwargs):
        """Returns neighbors of node"""
        neighs = self._overlay.neighbors(self)
        return self._overlay.filter(neighs, *args, **kwargs)

    def neighbor_interfaces(self, *args, **kwargs):
        #TODO: implement filtering for args and kwargs
        if len(args) or len(kwargs):
            log.warning("Attribute-based filtering not currently supported" 
                    " for neighbor_interfaces")

        return iter(edge.dst_int for edge in self.edges())


    @property
    def label(self):
        """Returns node label (mapped from ANM)"""
        return self.__repr__()

    @property
    def phy(self):
        """Shortcut back to physical OverlayNode
        Same as node.overlay.phy
        ie node.phy.x is same as node.overlay.phy.x
        """
# refer back to the physical node, to access attributes such as name
        return OverlayNode(self.anm, "phy", self.node_id)

    def dump(self):
        """Dump attributes of this node"""
        return str(self._graph.node[self.node_id])

    def edges(self, *args, **kwargs):
        """Edges to/from this node"""
        return self._overlay.edges(self, *args, **kwargs)

    def __repr__(self):
        """Try label if set in overlay, otherwise from physical,
        otherwise node id"""
        try:
            return self.anm.node_label(self)
        except KeyError:
            try:
                return self._graph.node[self.node_id]['label']
            except KeyError:
                return self.node_id  # node not in physical graph

    def __getattr__(self, key):
        """Returns node property
        This is useful for accesing attributes passed through from graphml"""
        try:
            return self._graph.node[self.node_id].get(key)
        except KeyError:
            return

    def get(self, key):
        """For consistency, node.get(key) is neater than getattr(node, key)"""
        return getattr(self, key)

    def __setattr__(self, key, val):
        """Sets node property
        This is useful for accesing attributes passed through from graphml"""
        try:
            self._graph.node[self.node_id][key] = val
        except KeyError:
            self._graph.add_node(self.node_id)
            self.set(key, val)

    def set(self, key, val):
        """For consistency, node.set(key, value) is neater 
        than setattr(node, key, value)"""
        return self.__setattr__(key, val)


@functools.total_ordering
class OverlayEdge(object):
    """API to access link in network"""
    def __init__(self, anm, overlay_id, src_id, dst_id):
# Set using this method to bypass __setattr__
        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
        object.__setattr__(self, 'src_id', src_id)
        object.__setattr__(self, 'dst_id', dst_id)

    def __eq__(self, other):
        """"""
        try:
            return (self.src_id, self.dst_id) == (other.src_id, other.dst_id)
        except AttributeError:
            return self.node_id == other

    def __repr__(self):
        """String of node"""
        return "%s: (%s, %s)" % (self.overlay_id, self.src, self.dst)

    def __getnewargs__(self):
        """"""
        return ()

    def __getitem__(self, key):
        """"""
        overlay = OverlayGraph(self.anm, key)
        return overlay.edge(self)

    def __getstate__(self):
        """For pickling"""
        return (self.anm, self.overlay_id, self.src_id, self.dst_id)

    def __lt__(self, other):
        """"""
        return ((self.src.node_id, self.dst.node_id) < (other.src.node_id, 
            other.dst.node_id))

    def __setstate__(self, state):
        """For pickling"""
        self._overlays = state
        (anm, overlay_id, src_id, dst_id) = state
        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
        object.__setattr__(self, 'src_id', src_id)
        object.__setattr__(self, 'dst_id', dst_id)

    @property
    def src(self):
        """Source node of edge"""
        return OverlayNode(self.anm, self.overlay_id, self.src_id)

    @property
    def dst(self):
        """Destination node of edge"""
        return OverlayNode(self.anm, self.overlay_id, self.dst_id)

    def apply_to_interfaces(self, attribute):
        val = self.__getattr__(attribute)
        self.src_int.__setattr__(attribute, val)
        self.dst_int.__setattr__(attribute, val)

    @property
    def src_int(self):
        """Interface bound to source node of edge"""
        src_int_id = self._interfaces[self.src_id]
        return overlay_interface(self.anm, self.overlay_id, self.src_id, src_int_id) 

    @property
    def dst_int(self):
        """Interface bound to destination node of edge"""
        dst_int_id = self._interfaces[self.dst_id]
        return overlay_interface(self.anm, self.overlay_id, self.dst_id, dst_int_id) 

    def attr_equal(self, *args):
        """Return edges which both src and dst have attributes equal"""
        return all(getattr(self.src, key) == getattr(self.dst, key) 
                for key in args)

    def attr_both(self, *args):
        """Return edges which both src and dst have attributes set"""
        return all((getattr(self.src, key) and getattr(self.dst, key)) 
                for key in args)

    def attr_any(self, *args):
        """Return edges which either src and dst have attributes set"""
        return all((getattr(self.src, key) or getattr(self.dst, key))
                for key in args)

    def dump(self):
        return str(self._graph[self.src_id][self.dst_id])

    def __nonzero__(self):
        """Allows for checking if edge exists
        """
        return self._graph.has_edge(self.src_id, self.dst_id)

    def bind_interface(self, node, interface):
        """Bind this edge to specified index"""
        self._interfaces[node.id] = interface

    def interfaces(self):
        #TODO: warn if interface doesn't exist on node
        return iter(overlay_interface(self.anm, self.overlay_id, node_id, interface_id) 
                for (node_id, interface_id) in self._interfaces.items())

    @property
    def _graph(self):
        """Return graph the node belongs to"""
        return self.anm.overlay_nx_graphs[self.overlay_id]

    def get(self, key):
        """For consistency, edge.get(key) is neater than getattr(edge, key)"""
        return self.__getattr__(key)

    def set(self, key, val):
        """For consistency, edge.set(key, value) is neater than 
        setattr(edge, key, value)"""
        return self.__setattr__(key, val)

    def __getattr__(self, key):
        """Returns edge property"""
        return self._graph[self.src_id][self.dst_id].get(key)

    def __setattr__(self, key, val):
        """Sets edge property"""
        self._graph[self.src_id][self.dst_id][key] = val


class OverlayGraphData(object):
    """API to access link in network"""
    def __init__(self, anm, overlay_id):
# Set using this method to bypass __setattr__
        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)

    def __repr__(self):
        """"""
        return "Data for (%s, %s)" % (self.anm, self.overlay_id)

    def dump(self):
        """"""
        print str(self._graph.graph)

    @property
    def _graph(self):
        # access underlying graph for this OverlayNode
        return self.anm.overlay_nx_graphs[self.overlay_id]

    def __getattr__(self, key):
        return self._graph.graph.get(key)

    def __setattr__(self, key, val):
        self._graph.graph[key] = val

    def __getitem__(self, key):
        """"""
        return self._graph.graph.get(key)

    def __setitem__(self, key, val):
        """"""
        self._graph.graph[key] = val


class OverlayBase(object):
    """Base class for overlays - overlay graphs, subgraphs, projections, etc"""

    def __init__(self, anm, overlay_id):
        """"""
        if overlay_id not in anm.overlay_nx_graphs:
            raise OverlayNotFound(overlay_id)
        self._anm = anm
        self._overlay_id = overlay_id

    def __repr__(self):
        """"""
        return self._overlay_id

    @property
    def data(self):
        """Returns data stored on this overlay graph"""
        return OverlayGraphData(self._anm, self._overlay_id)

    def __contains__(self, n):
        """"""
        return n.node_id in self._graph

    def interface(self, interface):
        """"""
        return overlay_interface(self._anm, self._overlay_id,
                interface.node_id, interface.interface_id)

    def edge(self, edge_to_find, dst_to_find=None):
        """returns edge in this graph with same src and same edge_id"""
        try:
            src = edge_to_find
            dst = dst_to_find
            src.lower()
            dst.lower()
            if self._graph.has_edge(src, dst):
                return OverlayEdge(self._anm, self._overlay_id, src, dst)
        except AttributeError:
            pass # not strings
        except TypeError:
            pass

        try:
            if dst_to_find:
                src_id = edge_to_find.node_id
                search_id = dst_to_find.node_id
            else:
                src_id = edge_to_find.src_id
                search_id = edge_to_find.edge_id
        except AttributeError:
            src_id = None
            search_id = edge_to_find

        for src, dst in self._graph.edges_iter(src_id):
            try:
                if self._graph[src][dst]['edge_id'] == search_id:
                    return OverlayEdge(self._anm, self._overlay_id, src, dst)
                elif (src, dst) == (src_id, search_id): 
                    # searching by nodes
                    return OverlayEdge(self._anm, self._overlay_id, src, dst)
            except KeyError:
                pass  # no edge_id for this edge

    def __getitem__(self, key):
        """"""
        return self.node(key)

    def node(self, key):
        """Returns node based on name
        This is currently O(N). Could use a lookup table"""
        try:
            if key.node_id in self._graph:
                return OverlayNode(self._anm, self._overlay_id, key.node_id)
        except AttributeError:
            # doesn't have node_id, likely a label string, search based on this
            # label
            for node in self:
                if str(node) == key:
                    return node
            log.warning("Unable to find node %s in %s " % (key, self))
            return None

    def degree(self, node):
        """"""
        return node.degree()

    def neighbors(self, node):
        return iter(OverlayNode(self._anm, self._overlay_id, node)
                    for node in self._graph.neighbors(node.node_id))

    def overlay(self, key):
        """Get to other overlay graphs in functions"""
        return OverlayGraph(self._anm, key)

    @property
    def name(self):
        """"""
        return self.__repr__()

    def node_label(self, node):
        """"""
        return repr(OverlayNode(self._anm, self._overlay_id, node))

    def dump(self):
        """"""
        self._anm.dump_graph(self)

    def has_edge(self, edge):
        """Tests if edge in graph"""
        return self._graph.has_edge(edge.src, edge.dst)

    def __iter__(self):
        """"""
        return iter(OverlayNode(self._anm, self._overlay_id, node)
                    for node in self._graph)

    def __len__(self):
        """"""
        return len(self._graph)

    def nodes(self, *args, **kwargs):
        """"""
        result = self.__iter__()
        if len(args) or len(kwargs):
            result = self.filter(result, *args, **kwargs)
        return result

    def routers(self, *args, **kwargs):
        """Shortcut for nodes(), sets device_type to be router"""
        kwargs['device_type'] = 'router'
        return self.nodes(*args, **kwargs)

    def device(self, key):
        """To access programatically"""
        return OverlayNode(self._anm, self._overlay_id, key)

    def groupby(self, attribute, nodes=None):
        """Returns a dictionary sorted by attribute

        >>> G_in.groupby("asn")
        {u'1': [r1, r2, r3, sw1], u'2': [r4]}
        """
        result = {}

        if not nodes:
            data = self.nodes()
        else:
            data = nodes
        data = sorted(data, key=lambda x: x.get(attribute))
        for key, grouping in itertools.groupby(data,
                key=lambda x: x.get(attribute)):
            result[key] = list(grouping)

        return result

    def filter(self, nbunch=None, *args, **kwargs):
        """"""
        if not nbunch:
            nbunch = self.nodes()

        def filter_func(node):
            """Filter based on args and kwargs"""
            return (
                all(getattr(node, key) for key in args) and
                all(getattr(
                    node, key) == val for key, val in kwargs.items())
            )

        return (n for n in nbunch if filter_func(n))

    def edges(self, src_nbunch=None, dst_nbunch=None, *args, **kwargs):
        """"""
# nbunch may be single node
        if src_nbunch:
            try:
                src_nbunch = src_nbunch.node_id
            except AttributeError:
                src_nbunch = (n.node_id for n in src_nbunch)
                              # only store the id in overlay

        def filter_func(edge):
            """Filter based on args and kwargs"""
            return (
                all(getattr(edge, key) for key in args) and
                all(getattr(
                    edge, key) == val for key, val in kwargs.items())
            )

        valid_edges = (
            (src, dst) for (src, dst) in self._graph.edges_iter(src_nbunch))
        if dst_nbunch:
            try:
                dst_nbunch = dst_nbunch.node_id
                dst_nbunch = set([dst_nbunch])
                                 # faster membership test than other sequences
            except AttributeError:
                dst_nbunch = (n.node_id for n in dst_nbunch)
                              # only store the id in OverlayEdge
                dst_nbunch = set(
                    dst_nbunch)  # faster membership test than other sequences

            valid_edges = ((src, dst) for (src, dst) in valid_edges
                           if dst in dst_nbunch)

        if len(args) or len(kwargs):
            all_edges = iter(OverlayEdge(self._anm, self._overlay_id, src, dst)
                             for (src, dst) in valid_edges)
            result = (edge for edge in all_edges if filter_func(edge))
        else:
            result = (OverlayEdge(self._anm, self._overlay_id, src, dst)
                      for (src, dst) in valid_edges)
        return result


class OverlaySubgraph(OverlayBase):
    """OverlaySubgraph"""

    def __init__(self, anm, overlay_id, graph, name=None):
        """"""
        super(OverlaySubgraph, self).__init__(anm, overlay_id)
        self._graph = graph
        self._subgraph_name = name

    def __repr__(self):
        return self._subgraph_name or "subgraph"


class OverlayGraph(OverlayBase):
    """API to interact with an overlay graph in ANM"""

    @property
    def anm(self):
        """Returns anm for this overlay"""
        return self._anm

    @property
    def _graph(self):
        """Access underlying graph for this OverlayNode"""
        return self._anm.overlay_nx_graphs[self._overlay_id]

    def _replace_graph(self, graph):
        """"""
        self._anm.overlay_nx_graphs[self._overlay_id] = graph

    # these work similar to their nx counterparts: just need to strip the
    # node_id
    def add_nodes_from(self, nbunch, retain=None, update=False, **kwargs):
        """Update won't append data (which could clobber) if node exists"""
        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        if not update:
# filter out existing nodes
            nbunch = (n for n in nbunch if n not in self._graph)

        nbunch = list(nbunch)
        node_ids = list(nbunch)  # before appending retain data

        if len(retain):
            add_nodes = []
            for node in nbunch:
                data = dict((key, node.get(key)) for key in retain)
                add_nodes.append((node.node_id, data))
            nbunch = add_nodes
        else:
            nbunch = (
                n.node_id for n in nbunch)  # only store the id in overlay

        self._graph.add_nodes_from(nbunch, **kwargs)
        self._init_interfaces(node_ids)

    def add_node(self, node, retain=None, **kwargs):
        """Adds node to overlay"""
        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        try:
            node_id = node.id
        except AttributeError:
            node_id = node  # use the string node id

        data = {}
        if len(retain):
            data = dict((key, node.get(key)) for key in retain)
            kwargs.update(data)  # also use the retained data
        self._graph.add_node(node_id, kwargs)
        self._init_interfaces([node_id])

    def _init_interfaces(self, nbunch=None):
        """Initialises interfaces"""
        if not nbunch:
            nbunch = [n for n in self._graph.nodes()]

        try:
            nbunch = list(unwrap_nodes(nbunch))
        except AttributeError:
            pass  # don't need to unwrap

        phy_graph = self._anm.overlay_nx_graphs["phy"]

        for node in nbunch:
            try:
                phy_interfaces = phy_graph.node[node]["_interfaces"]
                interface_data = {
                    'description': None,
                    'type': 'physical',
                }
                # need to do dict() to copy, otherwise all point to same memory location -> clobber
                data = dict(
                    (key, dict(interface_data)) for key in phy_interfaces)
                self._graph.node[node]['_interfaces'] = data
            except KeyError:
# no counterpart in physical graph, initialise
                log.debug("Initialise interfaces for %s in %s" % (
                    node, self._overlay_id))
                self._graph.node[node]['_interfaces'] = {0:
                                                         {'description': 'loopback',
                                                             'type': 'loopback'}}

    def allocate_interfaces(self):
        """allocates edges to interfaces"""
        # int_counter = (n for n in itertools.count() if n not in
        self._init_interfaces()

        def numeric_id(edge):
            try:
                return int(edge.edge_id.split("_")[0])
            except ValueError:
                return edge # can't casr to int

        ebunch = sorted(self.edges(), key = numeric_id)

        for edge in ebunch:
            src = edge.src
            dst = edge.dst
            src_int_id = src._add_interface('%s to %s' % (src, dst))
            dst_int_id = dst._add_interface('%s to %s' % (dst, src))
            edge._interfaces = {}
            edge._interfaces[src.id] = src_int_id
            edge._interfaces[dst.id] = dst_int_id

    def __delitem__(self, key):
        """Alias for remove_node. Allows
        >>> del overlay[node]
        """
        self.remove_node(key)

    def remove_node(self, node):
        """Removes a node from the overlay"""
        try:
            node_id = node.node_id
        except AttributeError:
            node_id = node
        self._graph.remove_node(node_id)

    def add_edge(self, src, dst, retain=None, **kwargs):
        """Adds an edge to the overlay"""
        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list
        self.add_edges_from([(src, dst)], retain, **kwargs)


    def remove_edges_from(self, ebunch):
        """Removes set of edges from ebunch"""
        try:
            ebunch = unwrap_edges(ebunch)
        except AttributeError:
            pass  # don't need to unwrap
        self._graph.remove_edges_from(ebunch)

    def add_edges(self, *args, **kwargs):
        """Adds a set of edges. Alias for add_edges_from"""
        self.add_edges_from(args, kwargs)

    def add_edges_from(self, ebunch, bidirectional=False,
            retain=None, **kwargs):
        """Add edges. Unlike NetworkX, can only add an edge if both 
        src and dst in graph already.
        If they are not, then they will not be added (silently ignored)

        Bidirectional will add edge in both directions. Useful if going 
        from an undirected graph to a
        directed, eg G_in to G_bgp
        """
        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        retain.append("edge_id")
        retain.append("_interfaces")
        try:
            if len(retain):
                add_edges = []
                for edge in ebunch:
                    data = dict((key, edge.get(key)) for key in retain)
                    add_edges.append(
                        (edge.src.node_id, edge.dst.node_id, data))
                ebunch = add_edges
            else:
                ebunch = [(e.src.node_id, e.dst.node_id, {}) for e in ebunch]
        except AttributeError:
            ebunch = [(src.node_id, dst.node_id, {"_interfaces": {}}) for src, dst in ebunch]

        ebunch = [(src, dst, data) for (src, dst, data)
                  in ebunch if src in self._graph and dst in self._graph]
        if bidirectional:
            ebunch += [(dst, src, data) for (
                src, dst, data) in ebunch if src in self._graph 
                and dst in self._graph]

        self._graph.add_edges_from(ebunch, **kwargs)

    def update(self, nbunch=None, **kwargs):
        """Sets property defined in kwargs to all nodes in nbunch"""
        if nbunch is None:
            nbunch = self.nodes()
        for node in nbunch:
            for key, value in kwargs.items():
                node.set(key, value)

    def update_edges(self, ebunch=None, **kwargs):
        """Sets property defined in kwargs to all edges in ebunch"""
        if not ebunch:
            ebunch = self.edges()
        for edge in ebunch:
            for key, value in kwargs.items():
                edge.set(key, value)

    def subgraph(self, nbunch, name=None):
        """"""
        nbunch = (n.node_id for n in nbunch)  # only store the id in overlay
        return OverlaySubgraph(self._anm, self._overlay_id, 
                self._graph.subgraph(nbunch), name)


class AbstractNetworkModel(object):
    """"""
    def __init__(self):
        """"""
        self._overlays = {}
        self.add_overlay("phy")
        self.add_overlay("graphics")

        self.label_seperator = "_"
        self.label_attrs = ['label']
        self._build_node_label()
        self.timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())

    def __repr__(self):
        """"""
        return "ANM %s" % self.timestamp

    @staticmethod
    def __getnewargs__():
        """"""
        return ()

    def __getstate__(self):
        """For pickling"""
        return (self._overlays, self.label_seperator, self.label_attrs)

    @property
    def overlay_nx_graphs(self):
        """"""
        return self._overlays

    def has_overlay(self, overlay_id):
        """"""
        return overlay_id in self._overlays

    def __setstate__(self, state):
        """For pickling"""
        (overlays, label_seperator, label_attrs) = state
        self._overlays = overlays
        self.label_seperator = label_seperator
        self.label_attrs = label_attrs
        self._build_node_label()

    def save(self):
        """"""
        import autonetkit.ank_json as ank_json
        import os
        import gzip
        archive_dir = os.path.join("versions", "anm")
        if not os.path.isdir(archive_dir):
            os.makedirs(archive_dir)

        data = ank_json.jsonify_anm(self)
        json_file = "anm_%s.json.gz" % self.timestamp
        json_path = os.path.join(archive_dir, json_file)
        log.debug("Saving to %s" % json_path)
        with gzip.open(json_path, "wb") as json_fh:
            json_fh.write(data)

    def restore_latest(self, directory=None):
        """Restores latest saved ANM"""
        import os
        import glob
        if not directory:
            directory = os.path.join("versions", "anm")

        glob_dir = os.path.join(directory, "*.json.gz")
        pickle_files = glob.glob(glob_dir)
        pickle_files = sorted(pickle_files)
        try:
            latest_file = pickle_files[-1]
        except IndexError:
# No files loaded
            log.warning("No previous ANM saved. Please compile new ANM")
            return
        self.restore(latest_file)

    def restore(self, pickle_file):
        """"""
        import json
        import gzip
        import autonetkit.ank_json as ank_json
        log.debug("Restoring %s" % pickle_file)
        with gzip.open(pickle_file, "r") as fh:
            data = json.load(fh)
            for overlay_id, graph_data in data.items():
                self._overlays[
                    overlay_id] = ank_json.ank_json_loads(graph_data)

    @property
    def _phy(self):
        """"""
        return OverlayGraph(self, "phy")

    def initialise_graph(self, graph):
        """Sets input graph. Converts to undirected. 
        Initialises graphics overlay."""
        graph = nx.Graph(graph)
        g_in = self.add_overlay("input", graph=graph, directed=False)
        g_graphics = self['graphics']
        g_graphics.add_nodes_from(g_in, retain=['x', 'y', 'device_type',
                                                'device_subtype', 'pop', 
                                                'asn'])

        return g_in

    def add_overlay(self, name, nodes=None, graph=None, directed=False,
            multi_edge=False, retain=None):
        """Adds overlay graph of name name"""
        if graph:
            if not directed and graph.is_directed():
                log.info("Converting graph %s to undirected" % name)
                graph = nx.Graph(graph)

        elif directed:
            if multi_edge:
                graph = nx.MultiDiGraph()
            else:
                graph = nx.DiGraph()
        else:
            if multi_edge:
                graph = nx.MultiGraph()
            else:
                graph = nx.Graph()

        self._overlays[name] = graph
        overlay = OverlayGraph(self, name)
        overlay.allocate_interfaces()
        if nodes:
            retain = retain or []  # default is an empty list
            overlay.add_nodes_from(nodes, retain)
        return overlay

    def overlays(self):
        """"""
        return self._overlays.keys()

    def devices(self, *args, **kwargs):
        """"""
        return self._phy.filter(*args, **kwargs)

    def __getitem__(self, key):
        """"""
        return OverlayGraph(self, key)

    def node_label(self, node):
        """Returns node label from physical graph"""
        return self.default_node_label(node)

    def _build_node_label(self):
        """"""
        def custom_label(node):
            return self.label_seperator.join(
                    str(self._overlays['phy'].node[node.node_id].get(val))
                    for val in self.label_attrs
                    if self._overlays['phy'].node[node.node_id].get(val)
                    is not None)

        self.node_label = custom_label

    def set_node_label(self, seperator, label_attrs):
        """"""
        try:
            label_attrs.lower()
            label_attrs = [label_attrs]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        self.label_seperator = seperator
        self.label_attrs = label_attrs

    def dump_graph(self, graph):
        """"""
        print "----Graph %s----" % graph
        print "Graph"
        print self.dump_graph_data(graph)
        print "Nodes"
        print self.dump_nodes(graph)
        print "Edges"
        print self.dump_edges(graph)

    @staticmethod
    def dump_graph_data(graph):
        """"""
        debug_data = dict((key, val)
                          for key, val in sorted(graph._graph.graph.items()))
        return pprint.pformat(debug_data)

    @staticmethod
    def dump_nodes(graph):
        """"""
        debug_data = dict((graph.node_label(node), data)
                          for node, data in (graph._graph.nodes(data=True)))
        return pprint.pformat(debug_data)

    @staticmethod
    def dump_edges(graph):
        """"""
        debug_data = dict(((graph.node_label(src), graph.node_label(dst)),
            data) for src, dst, data in (graph._graph.edges(data=True)))
        return pprint.pformat(debug_data)



import autonetkit.log as log
import pkg_resources
try:
    import textfsm
except ImportError:
    log.info("Unable to import TextFSM")

def traceroute(host, nidb, data):
    template_file = pkg_resources.resource_filename(__name__, "../textfsm/linux/traceroute")
    template = open(template_file)
    re_table = textfsm.TextFSM(template)
    routes = re_table.ParseText(data)
    #print "\t".join(re_table.header)
    for route in routes:
        #print "\t".join(route)
        route = route[0] # first element of table: todo make this programatic from table.header
        #print reverse_lookup(nidb, route)

    try:
        return [reverse_lookup(nidb, route[0])[1] for route in routes]
    except TypeError:
        log.info("Unable to parse %s" % routes)
        return []


def reverse_lookup(nidb, address):
    for node in nidb.nodes("is_l3device"):
        if str(node.loopback) == address:
            return ("loopback", node)

        for interface in node.interfaces:
            if str(interface.ipv4_address) == address:
                return (interface.id, node)
            
    return None 

def reverse_tap_lookup(nidb, address):
    for node in nidb.nodes("is_l3device"):
        if str(node.tap.ip) == address:
            return node


def sh_ip_route(host, nidb, data, record_data = False):
    #TODO: Split this from being seperate module to being a workable function
    #TODO: need to copy across subnet to collision domain in nidb

    import pprint
    log.debug("Parsed input data %s" % data)

    template_file = pkg_resources.resource_filename(__name__, "../textfsm/quagga/sh_ip_route")
    template = open(template_file)
    re_table = textfsm.TextFSM(template)
    routes = re_table.ParseText(data)


    #pprint.pprint(routes)

    #with open("test_out.txt", "w") as fh:
        #fh.write(pprint.pformat(routes, width=200))


    subnets = {}
    loopbacks = {}
    infra_interfaces = {}


    for node in nidb:
        if node.collision_domain:
            subnets[str(node.ipv4_subnet)] = node
        else:
            loopbacks[str(node.loopback)] = node
            for interface in node.physical_interfaces:
                # find interface in nidb
                #nidb_interface = nidb.interface(interface)
                infra_interfaces[str(interface.ipv4_address)] = interface

    #pprint.pprint(subnets)

    """
    pprint.pprint(["%s: %s.%s" % (subnet, nidb.interface(i).id, i.node)
            for (subnet, i) in infra_interfaces.items()])
    """

    mapped_routes = []
    for route_entry in routes:
        try:
            route_entry[3] = subnets[route_entry[3]]
        except KeyError:
            continue

        try:
            route_entry[4] = infra_interfaces[route_entry[4]]
        except KeyError:
            continue

        mapped_routes.append((route_entry[3], route_entry[4]))


    start_node = host

    nodes = [r[0] for r in mapped_routes]
    nodes = [start_node]
    edges = []
    paths = [[start_node, r[1].node, r[0]] for r in mapped_routes]
    #pprint.pprint(paths)
    log.debug("Parsed paths %s" % paths)

    import autonetkit.ank_messaging as ank_messaging
    #ank_messaging.highlight(nodes, edges, paths)

    if record_data:
        import time
        import json
        paths_str = [[str(start_node), str(r[1].node), str(r[0])] for r in mapped_routes]
        json_file = "measured_%s.json" % time.strftime("%Y%m%d_%H%M%S", time.localtime())
        with open(json_file, "wb") as json_fh:
            json_fh.write(json.dumps(paths_str))

    return paths



from pyparsing import Suppress,Word,ZeroOrMore,alphas,nums,delimitedList, Literal, Group, Optional, Forward
import json

#TODO: see if can just use t for the functions rather than strc, loc, toks
#TODO: generate all of the "boilerplate" classes and match functions from YAML or other description
#TODO: consider the switch to lambda x: with __init__ rather than a function and a class
#TODO: add in parse fail handling for more descriptive error messages
#TODO: drop pol_clause.map, etc and use as children: or have visitor that returns these as children - easier to traverse tree

#TODO: make methods that atoms inherit: key, value eg (setLP, 120), (prefix_list, blah_abc)
class pol_conditional(object):
    pass

class pol_match_or_action(object):
    is_match = False
    is_then = False
    is_else = False

class pol_clause(pol_conditional):
    def __init__(self, match_clause = None, then_clause = None, else_clause = None):
        self.match_clause = match_clause
        self.then_clause = then_clause
        self.else_clause = else_clause

    def __repr__(self):
        if self.else_clause:
            return "If:\n%s\n%s\n%s" % (self.match_clause, self.then_clause, self.else_clause)
        return "If:\n %s\n %s" % (self.match_clause, self.then_clause)

class pol_then(pol_conditional):
    def __repr__(self):
        return "then"

class pol_else(pol_conditional):
    def __repr__(self):
        return "else"

def fn_if(strg, loc, toks):
# Extract matches and actions - need to search through list. Return first (only) instance
    toks = toks[0] # TODO: make this clearer - due to nesting of groups
    if toks[0].is_then:
        then_clause = toks[0]
        return pol_clause(None, then_clause, None)

    match_clause = [tok for tok in toks[0] if tok.is_match][0]
    then_clause = [tok for tok in toks[0] if tok.is_then][0]
    try:
        else_clause = [tok for tok in toks[0] if tok.is_else][0]
    except IndexError:
        else_clause = None
    return pol_clause(match_clause, then_clause, else_clause)

class pol_match(pol_match_or_action):
    def __init__(self, matches):
        self.is_match = True
        self.matches = matches

    def __repr__(self):
        return "Match: %s" % self.matches

class pol_match_atom(object):
    pass


class pol_match_tag(pol_match_atom):
    def __init__(self, tag):
        self.tag = tag

    def __repr__(self):
        return "tags contain %s" % self.tag

class pol_match_pl(pol_match_atom):
    def __init__(self, pl):
        self.pl = pl

    def __repr__(self):
        return "prefix_list %s" % self.pl

# Matches

def fn_match(strg, loc, toks):
    matches = toks[0]
    return pol_match(matches)

def fn_match_tags(strg, loc, toks):
    tag = toks[0][1]
    return pol_match_tag(tag)

def fn_match_pl(strg, loc, toks):
    pl = toks[0][2]
    return pol_match_pl(pl)

class pol_then(pol_match_or_action):
    def __init__(self, actions):
        self.is_then = True
        self.actions = actions

    def __repr__(self):
        return "Then: %s" % self.actions

class pol_else(pol_match_or_action):
    def __init__(self, actions):
        self.is_else = True
        self.actions = actions

    def __repr__(self):
        return "Else: %s" % self.actions

# Actions
class pol_action_atom(object):
    pass

class pol_set_lp(pol_action_atom):
    #TODO: take either an integer value, or if alpha, then need to add it to a list to perform ordering on... may be better to have two different actions, and then feed into the same class once processed.
    def __init__(self, pl):
        self.pl = pl

    def __repr__(self):
        return "setLP %s" % self.pl

class pol_set_med(pol_action_atom):
    def __init__(self, med):
        self.med = med

    def __repr__(self):
        return "setMED %s" % self.med

class pol_add_tag(pol_action_atom):
    def __init__(self, tag):
        self.tag = tag

    def __repr__(self):
        return "addTag %s" % self.tag

class pol_remove_tag(pol_action_atom):
    def __init__(self, tag):
        self.tag = tag

    def __repr__(self):
        return "removeTag %s" % self.tag

class pol_reject(pol_action_atom):
    def __init__(self):
        pass # no parameter for reject route

    def __repr__(self):
        return "reject" 

def fn_then(strg, loc, toks):
    actions = toks[0]
    return pol_then(actions)

def fn_standalone_then(strg, loc, toks):
    actions = toks[0]
    return pol_then(actions)

def fn_else(strg, loc, toks):
    actions = toks[0]
    return pol_else(actions)

def fn_set_lp(strg, loc, toks):
    lp = toks[0][1]
    return pol_set_lp(lp)

def fn_set_med(strg, loc, toks):
    med = toks[0][1]
    return pol_set_med(med)

def fn_add_tag(strg, loc, toks):
    tag = toks[0][1]
    return pol_add_tag(tag)

def fn_remove_tag(strg, loc, toks):
    tag = toks[0][1]
    return pol_remove_tag(tag)

def fn_reject(strg, loc, toks):
    return pol_reject()

and_token = Suppress("and")
is_token = Literal("is")

# matches
# tags contain xyz
match_tags = Group(Literal("tags contain") + Word(alphas)).setParseAction(fn_match_tags)
match_pl = Group(Literal("prefix_list") + is_token + Word(alphas)).setParseAction(fn_match_pl)

# advanced matches (need to be expanded)
#TODO: expand these as match... and use memoize decorator to cache the search query
# origin(...)
# transit(...)

match_clause = (match_tags | match_pl)
match_clauses = Group(Suppress("(") + match_clause + ZeroOrMore(and_token + match_clause) 
        + Suppress(")")).setParseAction(fn_match)

# actions
# setLP lp
token_set_lp = Group(Literal("setLP") + Word(nums)).setParseAction(fn_set_lp)
# setMED med
token_set_med = Group(Literal("setMED") + Word(nums)).setParseAction(fn_set_med)
# addTag tag
token_add_tag = Group(Literal("addTag") + Word(nums)).setParseAction(fn_add_tag)
# removeTag tag
token_remove_tag = Group(Literal("removeTag") + Word(nums)).setParseAction(fn_remove_tag)
# reject
token_reject = Literal("reject").setParseAction(fn_reject)

# these can either take an IP or a host - todo: do we also want to allow specifying interface?
# setNextHop ip | host
# setOriginAttribute ip | host

action_clause = (token_set_lp | token_set_med | token_add_tag | token_remove_tag | token_reject ) 
then_actions =  action_clause + ZeroOrMore(and_token + action_clause)
then_clause = Group(Suppress("then") + Suppress("(") + then_actions 
        + Suppress(")")).setParseAction(fn_then)
standlone_then_clause = Group(then_actions).setParseAction(fn_standalone_then)

bgp_pol = Forward() # if clause can be present inside an else

else_actions = (action_clause + ZeroOrMore(and_token + action_clause)) | bgp_pol
else_clause = Group(Suppress("else") + Suppress("(") + else_actions 
        + Suppress(")")).setParseAction(fn_else)

bgp_pol << Group(Group(Suppress("if") + match_clauses + then_clause + Optional(else_clause)) | standlone_then_clause).setParseAction(fn_if)


class BgpPolEncoder(json.JSONEncoder):
    """Recursive handling and formatting for BGP Policy export in JSON format"""
    def default(self, obj):
        if isinstance(obj, pol_clause):
            #TODO: add documentation about serializing anm nodes
            return {'match': obj.match_clause, 'then': obj.then_clause, 'else': obj.else_clause}

        if isinstance(obj, pol_match):
            return [match for match in obj.matches]
        if isinstance(obj, pol_then):
            return [action for action in obj.actions]
        if isinstance(obj, pol_else):
            return [action for action in obj.actions]

        if isinstance(obj, pol_match_atom):
            return str(obj)
        if isinstance(obj, pol_action_atom):
            return str(obj)

        print "objs is", type(obj)
        return json.JSONEncoder.default(self, obj)

policies = [
"if (tags contain aaa and prefix_list is zyx and tags contain bbb) then (reject and setLP 100)",
"if (tags contain aaa and prefix_list is zyx and tags contain bbb) then (reject and setLP 100) else (setLP 200)",
"if (prefix_list is ccc) then (reject and setLP 100) else (reject and setMED 240 and setLP 210)",
("if (prefix_list is ccc) then (reject and setLP 100) else "
"(if (prefix_list is ccc) then (setLP 120) else (setMED 230))"),
"if (prefix_list is ccc) then (setLP 120) else (setMED 230))",
"reject and setLP 200",
]


for my_policy in policies:
    print "Policy: ", my_policy
    results = bgp_pol.parseString(my_policy)
    #print "results", results
    results = list(results) # list context
    data = json.dumps(results, cls=BgpPolEncoder, indent = 4)
    print data

# also need to walk tree to build up a list of all tags present, and all prefix lists

# end up with lists of named tuples for tokens: (match, value), and (action, value) that can then format inside the template - that way slight differences between syntaxes are able to be handles in templates rather than back up in the parser

#TODO: implement both visitors and transformers

#TODO: map the prefix_list and tags to lists for allocation, same as in old version
#TODO: could this be done by a variation of the BgpPolEncoder visitor, in that any instances of prefix_list or tag types are returned recursively and then flattened at the end?



import autonetkit.ank as ank
import itertools
import netaddr
import os
from collections import defaultdict
import string
from datetime import datetime
import autonetkit.log as log
import autonetkit.plugins.naming as naming
import autonetkit.config
from autonetkit.ank_utils import alphabetical_sort as alpha_sort
from autonetkit.ank import sn_preflen_to_network

def dot_to_underscore(instring):
    """Replace dots with underscores"""
    return instring.replace(".", "_")

class RouterCompiler(object):
    """Base router compiler"""
    lo_interface = "lo0"
    lo_interface_prefix = "lo"
# and set per platform

    def ibgp_session_data(self, session, ip_version):
        # Don't make staticmethod as may want to extend (e.g. in IOS compiler for vpnv4)
        node = session.src
        neigh = session.dst

        if ip_version == 4:
            neigh_ip = neigh['ipv4']
            use_ipv4 = True
            use_ipv6 = False
        elif ip_version == 6:
            neigh_ip = neigh['ipv6']
            use_ipv4 = False
            use_ipv6 = True

        data = {
            'neighbor': neigh.label,
            'use_ipv4': use_ipv4,
            'use_ipv6': use_ipv6,
            'asn': neigh.asn,
            'loopback': neigh_ip.loopback,
            # TODO: this is platform dependent???
            'update_source': node.loopback_zero.id,
            }
        return data

    def ebgp_session_data(self, session, ip_version):
        node = session.src
        neigh = session.dst
        if ip_version == 4:
            neigh_ip = neigh['ipv4']
            local_int_ip = session.src_int['ipv4'].ip_address
            dst_int_ip = session.dst_int['ipv4'].ip_address
            use_ipv4 = True
            use_ipv6 = False
        elif ip_version == 6:
            neigh_ip = neigh['ipv6']
            local_int_ip = session.src_int['ipv6'].ip_address
            dst_int_ip = session.dst_int['ipv6'].ip_address
            use_ipv4 = False
            use_ipv6 = True

        data = {
            'neighbor': neigh.label,
            'use_ipv4': use_ipv4,
            'use_ipv6': use_ipv6,
            'asn': neigh.asn,
            'loopback': neigh_ip.loopback,
            'local_int_ip': local_int_ip,
            'dst_int_ip': dst_int_ip,
            # TODO: change templates to access from node.bgp.lo_int
            'update_source': node.loopback_zero.id,
            }
        return data

    """Base Router compiler"""
    def __init__(self, nidb, anm):
        self.nidb = nidb
        self.anm = anm

    def compile(self, node):
        phy_node = self.anm['phy'].node(node)
        ipv4_node = self.anm['ipv4'].node(node)

        node.ip.use_ipv4 = phy_node.use_ipv4 or False
        node.ip.use_ipv6 = phy_node.use_ipv6 or False
        if not (node.ip.use_ipv4 and node.ip.use_ipv6):
            log.debug("Neither IPv4 nor IPv6 specified for %s, using IPv4" % node)
            node.ip.use_ipv4 = True

        node.label = naming.network_hostname(phy_node)
        node.input_label = phy_node.id
        node.loopback = ipv4_node.loopback
        node.loopback_subnet = netaddr.IPNetwork(node.loopback)
        node.loopback_subnet.prefixlen = 32

        self.interfaces(node)
        if node in self.anm['ospf']:
            self.ospf(node)
        if node in self.anm['isis']:
            self.isis(node)
        if node in self.anm['bgp']:
            self.bgp(node)

    def interfaces(self, node):
        node.interfaces = []

        node.loopback_zero.id = self.lo_interface
        node.loopback_zero.description = "Loopback"

        for interface in node.physical_interfaces:
            phy_int = self.anm['phy'].interface(interface)
            interface.physical = True
            #TODO: allocate ID in platform compiler

            if not phy_int:
                # for instance if added as management interface to nidb in compile
                continue
                
            interface.description = phy_int.description
            if node.ip.use_ipv4:
                ipv4_int = phy_int['ipv4']
                interface.ipv4_address = ipv4_int.ip_address
                interface.ipv4_subnet = ipv4_int.subnet
                interface.ipv4_cidr = sn_preflen_to_network(interface.ipv4_address,
                        interface.ipv4_subnet.prefixlen)

            if node.ip.use_ipv6:
                ipv6_int = phy_int['ipv6']
#TODO: for consistency, make ipv6_cidr
                interface.ipv6_subnet = ipv6_int.subnet
                interface.ipv6_address = sn_preflen_to_network(ipv6_int.ip_address,
                        interface.ipv6_subnet.prefixlen)

        for interface in node.loopback_interfaces:
            #TODO: check if nonzero is different to __eq__
            if interface == node.loopback_zero:
                continue
            else:
                #print "here for non zero", interface.id
                phy_int = self.anm['phy'].interface(interface)
                if node.ip.use_ipv4:
                    ipv4_int = phy_int['ipv4']
                    interface.ipv4_address = ipv4_int.loopback
                    interface.ipv4_subnet = node.loopback_subnet
                    interface.ipv4_cidr = sn_preflen_to_network(interface.ipv4_address,
                            interface.ipv4_subnet.prefixlen)

                if node.ip.use_ipv6:
                    ipv6_int = phy_int['ipv6']
#TODO: for consistency, make ipv6_cidr
                    #interface.ipv6_subnet = ipv6_int.loopback # TODO: do we need for consistency?
                    interface.ipv6_address = sn_preflen_to_network(
                            ipv6_int.loopback, 128)

                # secondary loopbacks
                #TODO: check why vrf names not showing up for all
                #print vrf_interface.vrf_name
                pass

            continue

    def ospf(self, node):
        """Returns OSPF links, also sets process_id
        """
        g_ospf = self.anm['ospf']
        g_ipv4 = self.anm['ipv4']

        node.ospf.loopback_area = g_ospf.node(node).area

        node.ospf.process_id = 1 #TODO: set this in build_network module
        node.ospf.lo_interface = self.lo_interface

        node.ospf.ospf_links = []
        added_networks = set()
        for interface in node.physical_interfaces:
            if interface.exclude_igp:
                continue # don't configure IGP for this interface
            ipv4_int = g_ipv4.interface(interface)
            ospf_int = g_ospf.interface(interface)
            if not ospf_int.is_bound:
                continue # not an OSPF interface
            try:
                ospf_cost = int(ospf_int.cost)
            except TypeError:
                try:
                    ospf_cost = netaddr.IPAddress(ospf_int.cost)
                except TypeError:
                    log.debug("Using default OSPF cost of 1 for %s on %s" % (ospf_int, node))
                    ospf_cost = 1 # default
            interface.ospf_cost = ospf_cost
            network = ipv4_int.subnet
            if (ospf_int and ospf_int.is_bound
                    and network not in added_networks):  # don't add more than once
                added_networks.add(network)
                node.ospf.ospf_links.append(
                    network=network,
                    area=ospf_int.area,
                )

    def bgp(self, node):
        phy_node = self.anm['phy'].node(node)
        g_bgp = self.anm['bgp']
        g_ipv4 = self.anm['ipv4']
        g_ipv6 = self.anm['ipv6']
        asn = phy_node.asn
        node.asn = asn
        node.bgp.ipv4_advertise_subnets = []
        if node.ip.use_ipv4:
            node.bgp.ipv4_advertise_subnets = g_ipv4.data.infra_blocks.get(
                asn) or []  # could be none (if one-node AS) default empty list
        node.bgp.ipv6_advertise_subnets = []
        if node.ip.use_ipv6:
            g_ipv6 = self.anm['ipv6']
            node.bgp.ipv6_advertise_subnets = g_ipv6.data.infra_blocks.get(
                asn) or []

        ibgp_neighbors = []
        ibgp_rr_clients = []
        ibgp_rr_parents = []

        g_ibgp_v4 = self.anm['ibgp_v4']
        for session in g_ibgp_v4.edges(phy_node):
            if session.exclude:
                log.debug("Skipping excluded ibgp session %s" % session)
                continue # exclude from regular ibgp config (eg for VRF, VPLS, etc)

            data = self.ibgp_session_data(session, ip_version = 4)

            direction = session.direction
            if direction == 'down':
                ibgp_rr_clients.append(data)
            elif direction == 'up':
                ibgp_rr_parents.append(data)
            else:
                ibgp_neighbors.append(data)

        #TODO: check v6 hierarchy only created if node set to being v4 or v6
        g_ibgp_v6 = self.anm['ibgp_v6']
        for session in g_ibgp_v6.edges(phy_node):
            if session.exclude:
                log.debug("Skipping excluded ibgp session %s" % session)
                continue # exclude from regular ibgp config (eg for VRF, VPLS, etc)
            data = self.ibgp_session_data(session, ip_version = 6)

            direction = session.direction
            if direction == 'down':
                ibgp_rr_clients.append(data)
            elif direction == 'up':
                ibgp_rr_parents.append(data)
            else:
                ibgp_neighbors.append(data)

        #TODO: update this to use ibgp_v4 and ibgp_v6 overlays

        node.bgp.ibgp_neighbors = ibgp_neighbors
        node.bgp.ibgp_rr_clients = ibgp_rr_clients 
        node.bgp.ibgp_rr_parents = ibgp_rr_parents

        #ebgp

        ebgp_neighbors = []
        g_ebgp_v4 = self.anm['ebgp_v4']
        for session in g_ebgp_v4.edges(phy_node):
            if session.exclude:
                log.debug("Skipping excluded ebgp session %s" % session)
                continue # exclude from regular ibgp config (eg for VRF, VPLS, etc)
            data = self.ebgp_session_data(session, ip_version = 4)
            ebgp_neighbors.append(data)

        g_ebgp_v6 = self.anm['ebgp_v6']
        for session in g_ebgp_v6.edges(phy_node):
            if session.exclude:
                log.debug("Skipping excluded ebgp session %s" % session)
                continue # exclude from regular ibgp config (eg for VRF, VPLS, etc)
            data = self.ebgp_session_data(session, ip_version = 6)
            ebgp_neighbors.append(data)

        node.bgp.ebgp_neighbors = ebgp_neighbors
        node.bgp.ebgp_neighbors.sort("asn")

        return

    def isis(self, node):
        g_isis = self.anm['isis']
        node.isis.isis_links = []

        for interface in node.physical_interfaces:
            if interface.exclude_igp:
                continue # don't configure IGP for this interface

            phy_int = self.anm['phy'].interface(interface)

            isis_int = phy_int['isis']
            if isis_int and isis_int.is_bound:
                isis_node = g_isis.node(node)
                interface.isis = {
                        'metric': isis_int.metric,
                        'process_id': node.isis.process_id,
                        'use_ipv4': node.ip.use_ipv4,
                        'use_ipv6': node.ip.use_ipv6,
                        'multipoint': isis_int.multipoint,
                        } #TODO: add wrapper for this

                node.isis.isis_links.append(
                        id = interface.id,
                        metric = isis_int.metric,
                        multipoint= isis_int.multipoint,
                        )

        g_isis = self.anm['isis']
        isis_node = self.anm['isis'].node(node)
        node.isis.net = isis_node.net
        node.isis.process_id = isis_node.process_id
        node.isis.lo_interface = self.lo_interface
# set isis on loopback_zero
        
        node.loopback_zero.isis = {
                        'use_ipv4': node.ip.use_ipv4,
                        'use_ipv6': node.ip.use_ipv6,
                        } #TODO: add wrapper for this

class QuaggaCompiler(RouterCompiler):
    """Base Quagga compiler"""
    lo_interface = "lo0:1"

    def compile(self, node):
        super(QuaggaCompiler, self).compile(node)


    def interfaces(self, node):
        """Quagga interface compiler"""
        ipv4_node = self.anm['ipv4'].node(node)
        phy_node = self.anm['phy'].node(node)

        super(QuaggaCompiler, self).interfaces(node)
        # OSPF cost

        if phy_node.is_l3device:
            node.loopback_zero.id = self.lo_interface
            node.loopback_zero.description = "Loopback"
            node.loopback_zero.ipv4_address=ipv4_node.loopback
            node.loopback_zero.ipv4_subnet=node.loopback_subnet

    def ospf(self, node):
        """Quagga ospf compiler"""
        super(QuaggaCompiler, self).ospf(node)

        # add eBGP link subnets
        node.ospf.passive_interfaces = []

        for interface in node.physical_interfaces:
            if interface.exclude_igp:
                continue # don't configure IGP for this interface

            bgp_int = self.anm['bgp'].interface(interface)
            if bgp_int.ebgp: # ebgp interface
                node.ospf.passive_interfaces.append(
                        id=interface.id,
                        )
                subnet = bgp_int['ipv4'].subnet
                default_ebgp_area = 0
                node.ospf.ospf_links.append(
                        network=subnet,
                        area=default_ebgp_area,
                        )

    def isis(self, node):
        """Sets ISIS links
        """
        g_isis = self.anm['isis']
        isis_node = g_isis.node(node)
        node.isis.net = isis_node.net
        node.isis.process_id = isis_node.process_id

# TODO: Don't render netkit lab topology if no netkit hosts

class IosBaseCompiler(RouterCompiler):
    """Base IOS compiler"""

    lo_interface_prefix = "Loopback"
    lo_interface = "%s%s" % (lo_interface_prefix, 0)

    def ibgp_session_data(self, session, ip_version):
        """Wraps RouterCompiler ibgp_session_data
        adds vpnv4 = True if ip_version == 4 and session is in g_ibgp_vpn_v4"""
        data = super(IosBaseCompiler, self).ibgp_session_data(session, ip_version)
        if ip_version == 4:
            g_ibgp_vpn_v4 = self.anm['ibgp_vpn_v4']
            if g_ibgp_vpn_v4.has_edge(session):
                data['use_vpnv4'] = True
        return data

    def compile(self, node):
        self.vrf_igp_interfaces(node)
        phy_node = self.anm['phy'].node(node)

        node.use_cdp = phy_node.use_cdp
        if phy_node.device_subtype == "os":
            # only copy across for certain reference platforms
            node.use_onepk = phy_node.use_onepk

        if node in self.anm['ospf']:
            node.ospf.use_ipv4 = phy_node.use_ipv4
            node.ospf.use_ipv6 = phy_node.use_ipv6

        if node in self.anm['isis']:
            node.isis.use_ipv4 = phy_node.use_ipv4
            node.isis.use_ipv6 = phy_node.use_ipv6

        super(IosBaseCompiler, self).compile(node)
        if node in self.anm['isis']:
            self.isis(node)

        node.label = self.anm['phy'].node(node).label
        """Note:
        VRFs are either: before, and mark IGP interfaces to skip;
        or after, and remove IGP interfaces
        """
        self.vrf(node)

    def interfaces(self, node):
        phy_loopback_zero = self.anm['phy'].interface(node.loopback_zero)
        if node.ip.use_ipv4:
            ipv4_loopback_subnet = netaddr.IPNetwork("0.0.0.0/32")
            ipv4_loopback_zero = phy_loopback_zero['ipv4']
            ipv4_address = ipv4_loopback_zero.ip_address
            node.loopback_zero.ipv4_address = ipv4_address
            node.loopback_zero.ipv4_subnet = ipv4_loopback_subnet
            node.loopback_zero.ipv4_cidr = sn_preflen_to_network(
                    ipv4_address, ipv4_loopback_subnet.prefixlen)

        if node.ip.use_ipv6:
            ipv6_loopback_zero = phy_loopback_zero['ipv6']
            node.loopback_zero.ipv6_address = sn_preflen_to_network(
                ipv6_loopback_zero.ip_address, 128)

        super(IosBaseCompiler, self).interfaces(node)

        for interface in node.physical_interfaces:
            interface.use_cdp = node.use_cdp # use node value

    def bgp(self, node):
        node.bgp.lo_interface = self.lo_interface
        super(IosBaseCompiler, self).bgp(node)

        # Only advertise loopbacks into eBGP
        if node.ip.use_ipv4:
            node.bgp.ipv4_advertise_subnets = [node.loopback_zero.ipv4_cidr]
        if node.ip.use_ipv6:
            node.bgp.ipv6_advertise_subnets = [node.loopback_zero.ipv6_address]

        # vrf
        #TODO: this should be inside vrf section?
        node.bgp.vrfs = []

        vrf_node = self.anm['vrf'].node(node)
        if vrf_node.vrf_role is "PE":

            # iBGP sessions for this VRF
            vrf_ibgp_neighbors = defaultdict(list)

            g_ibgp_v4 = self.anm['ibgp_v4']
            for session in g_ibgp_v4.edges(vrf_node):
                if session.exclude and session.vrf:
                    data = self.ibgp_session_data(session, ip_version = 4)
                    vrf_ibgp_neighbors[session.vrf].append(data)

            g_ibgp_v6 = self.anm['ibgp_v6']
            for session in g_ibgp_v6.edges(vrf_node):
                if session.exclude and session.vrf:
                    data = self.ibgp_session_data(session, ip_version = 6)
                    vrf_ibgp_neighbors[session.vrf].append(data)

            # eBGP sessions for this VRF
            vrf_ebgp_neighbors = defaultdict(list)

            g_ebgp_v4 = self.anm['ebgp_v4']
            for session in g_ebgp_v4.edges(vrf_node):
                if session.exclude and session.vrf:
                    data = self.ebgp_session_data(session, ip_version = 4)
                    vrf_ebgp_neighbors[session.vrf].append(data)

            g_ebgp_v6 = self.anm['ebgp_v6']
            for session in g_ebgp_v6.edges(vrf_node):
                if session.exclude and session.vrf:
                    data = self.ebgp_session_data(session, ip_version = 6)
                    vrf_ebgp_neighbors[session.vrf].append(data)

            for vrf in vrf_node.node_vrf_names:
                rd_index = vrf_node.rd_indices[vrf]
                rd = "%s:%s" % (node.asn, rd_index)
                node.bgp.vrfs.append(
                    vrf=vrf,
                    rd=rd,
                    use_ipv4=node.ip.use_ipv4,
                    use_ipv6=node.ip.use_ipv6,
                    vrf_ebgp_neighbors = vrf_ebgp_neighbors[vrf],
                    vrf_ibgp_neighbors = vrf_ibgp_neighbors[vrf],
                )

        # Retain route_target if in ibgp_vpn_v4 and RR or HRR (set in design)
        vpnv4_node = self.anm['ibgp_vpn_v4'].node(node)
        if vpnv4_node:
            retain = False
            if vpnv4_node.retain_route_target:
                retain = True
            node.bgp.vpnv4 = {'retain_route_target': retain}

    def vrf_igp_interfaces(self, node):
        # marks physical interfaces to exclude from IGP
        vrf_node = self.anm['vrf'].node(node)
        if vrf_node.vrf_role is "PE":
            for interface in node.physical_interfaces:
                vrf_int = self.anm['vrf'].interface(interface)
                if vrf_int.vrf_name:
                    interface.exclude_igp = True

    def vrf(self, node):
        g_vrf = self.anm['vrf']
        vrf_node = self.anm['vrf'].node(node)
        node.vrf.vrfs = []
        if vrf_node.vrf_role is "PE":
            #TODO: check if mpls ldp already set elsewhere
            for vrf in vrf_node.node_vrf_names:
                route_target = g_vrf.data.route_targets[node.asn][vrf]
                rd_index = vrf_node.rd_indices[vrf]
                rd = "%s:%s" % (node.asn, rd_index)

                node.vrf.vrfs.append({
                    'vrf': vrf,
                    "rd": rd,
                    'route_target': route_target,
                })

            for interface in node.interfaces:
                vrf_int = self.anm['vrf'].interface(interface)
                if vrf_int.vrf_name:
                    interface.vrf = vrf_int.vrf_name # mark interface as being part of vrf
                    if interface.physical:
                        interface.description += " vrf %s" % vrf_int.vrf_name

        if vrf_node.vrf_role in ("P", "PE"):
            # Add PE -> P, PE -> PE interfaces to MPLS LDP
            node.mpls.ldp_interfaces = []
            for interface in node.physical_interfaces:
                mpls_ldp_int = self.anm['mpls_ldp'].interface(interface)
                if mpls_ldp_int.is_bound:
                    node.mpls.ldp_interfaces.append(interface.id)
                    interface.use_mpls = True

        if vrf_node.vrf_role is "P":
            node.mpls.ldp_interfaces = []
            for interface in node.physical_interfaces:
                node.mpls.ldp_interfaces.append(interface.id)

        vrf_node = self.anm['vrf'].node(node)

        node.vrf.use_ipv4 = node.ip.use_ipv4
        node.vrf.use_ipv6 = node.ip.use_ipv6
        node.vrf.vrfs.sort("vrf")

        if node in self.anm['mpls_ldp']:
            node.mpls.enabled = True
            node.mpls.router_id = node.loopback_zero.id

    def ospf(self, node):
        super(IosBaseCompiler, self).ospf(node)
        for interface in node.physical_interfaces:
            phy_int = self.anm['phy'].interface(interface)

            ospf_int = phy_int['ospf']
            if ospf_int and ospf_int.is_bound:
                if interface.exclude_igp:
                    continue # don't configure IGP for this interface

                interface.ospf = {
                        'cost': ospf_int.cost,
                        'area': ospf_int.area,
                        'process_id': node.ospf.process_id,
                        'use_ipv4': node.ip.use_ipv4,
                        'use_ipv6': node.ip.use_ipv6,
                        'multipoint': ospf_int.multipoint,
                        } #TODO: add wrapper for this

class IosClassicCompiler(IosBaseCompiler):
    def compile(self, node):
        super(IosClassicCompiler, self).compile(node)

        phy_node = self.anm['phy'].node(node)
        if phy_node.include_csr:
            node.include_csr = True

    def bgp(self, node):
        super(IosClassicCompiler, self).bgp(node)

        vpnv4_neighbors = []
        if node.bgp.vpnv4:
            for neigh in node.bgp.ibgp_neighbors:
                if not neigh.use_ipv4:
                    continue
                #TODO: fix up limitation where can't add as overlay_data
                # (this causes problems serializing, when adding convert to dict?)
                neigh_data = dict(neigh)
                vpnv4_neighbors.append(neigh_data)

            for neigh in node.bgp.ibgp_rr_clients:
                if not neigh.use_ipv4:
                    continue
                neigh_data = dict(neigh)
                neigh_data['rr_client'] = True
                vpnv4_neighbors.append(neigh_data)

            for neigh in node.bgp.ibgp_rr_parents:
                if not neigh.use_ipv4:
                    continue
                neigh_data = dict(neigh)
                vpnv4_neighbors.append(neigh_data)

        node.bgp.vpnv4_neighbors = vpnv4_neighbors

class Ios2Compiler(IosBaseCompiler):

    def ospf(self, node):
        super(Ios2Compiler, self).ospf(node)
        g_ospf = self.anm['ospf']
        interfaces_by_area = defaultdict(list)

        for interface in node.physical_interfaces:
            if interface.exclude_igp:
                continue # don't configure IGP for this interface

            ospf_int = g_ospf.interface(interface)
            if ospf_int and ospf_int.is_bound:
                area = ospf_int.area
                area = str(area) # can't serialize IPAddress object to JSON
                interfaces_by_area[area].append({
                    'id': interface.id,
                    'cost': int(ospf_int.cost),
                    'passive': False,
                })

        loopback_zero = node.loopback_zero
        ospf_loopback_zero = g_ospf.interface(loopback_zero)
        router_area = ospf_loopback_zero.area # area assigned to router
        router_area = str(router_area) # can't serialize IPAddress object to JSON
        interfaces_by_area[router_area].append({
            'id': node.loopback_zero.id,
            'cost': 0,
            'passive': True,
        })


        node.ospf.interfaces = dict( interfaces_by_area)

class NxOsCompiler(IosBaseCompiler):
    def interfaces(self, node):
# need to aggregate areas
        super(NxOsCompiler, self).interfaces(node)

    def ospf(self, node):
        super(NxOsCompiler, self).ospf(node)
        # TODO: configure OSPF on loopback like example

# Platform compilers

class PlatformCompiler(object):
    """Base Platform Compiler"""
# and set properties in nidb._graph.graph
    def __init__(self, nidb, anm, host):
        self.nidb = nidb
        self.anm = anm
        self.host = host

    @property
    def timestamp(self):
        return self.nidb.timestamp

    def compile(self):
        # TODO: make this abstract
        pass


class JunosphereCompiler(PlatformCompiler):
    """Junosphere Platform Compiler"""
    def interface_ids(self):
        invalid = set([2])
        valid_ids = (x for x in itertools.count(0) if x not in invalid)
        for x in valid_ids:
            yield "ge-0/0/%s" % x

    def compile(self):
        log.info("Compiling Junosphere for %s" % self.host)
        g_phy = self.anm['phy']
        junos_compiler = JunosCompiler(self.nidb, self.anm)
        for phy_node in g_phy.nodes('is_router', host=self.host, syntax='junos'):
            nidb_node = self.nidb.node(phy_node)
            nidb_node.render.template = os.path.join("templates","junos.mako")
            nidb_node.render.dst_folder = os.path.join("rendered",self.host,"junosphere")
            nidb_node.render.dst_file = "%s.conf" % ank.name_folder_safe(
                phy_node.label)

            int_ids = self.interface_ids()
            for interface in nidb_node.physical_interfaces:
                interface.unit = 0
                interface.id = int_ids.next()

            junos_compiler.compile(nidb_node)


class NetkitCompiler(PlatformCompiler):
    """Netkit Platform Compiler"""
    @staticmethod
    def index_to_int_id(index):
        return "eth%s" % index

    def compile(self):
        log.info("Compiling Netkit for %s" % self.host)
        g_phy = self.anm['phy']
        quagga_compiler = QuaggaCompiler(self.nidb, self.anm)
# TODO: this should be all l3 devices not just routers
        for phy_node in g_phy.nodes('is_l3device', host=self.host, syntax='quagga'):
            folder_name = naming.network_hostname(phy_node)
            nidb_node = self.nidb.node(phy_node)
            nidb_node.render.base = os.path.join("templates","quagga")
            nidb_node.render.template = os.path.join("templates","netkit_startup.mako")
            nidb_node.render.dst_folder = os.path.join("rendered", self.host, "netkit")
            nidb_node.render.base_dst_folder = os.path.join("rendered", self.host, "netkit", folder_name)
            nidb_node.render.dst_file = "%s.startup" % folder_name

# allocate zebra information
            if nidb_node.is_router:
                nidb_node.zebra.password = "1234"
            hostname = folder_name
            if hostname[0] in string.digits:
                hostname = "r" + hostname
            nidb_node.hostname = hostname  # can't have . in quagga hostnames
            nidb_node.ssh.use_key = True  # TODO: make this set based on presence of key

            # Note this could take external data
            int_ids = itertools.count(0)
            for interface in nidb_node.physical_interfaces:
                numeric_id = int_ids.next()
                interface.numeric_id = numeric_id
                interface.id = self.index_to_int_id(numeric_id)

# and allocate tap interface
            nidb_node.tap.id = self.index_to_int_id(int_ids.next())

            quagga_compiler.compile(nidb_node)

            # TODO: move these into inherited BGP config
            if nidb_node.bgp:
                nidb_node.bgp.debug = True
                static_routes = []
                nidb_node.zebra.static_routes = static_routes

        # and lab.conf
        self.allocate_tap_ips()
        self.lab_topology()

    def allocate_tap_ips(self):
        settings = autonetkit.config.settings
        # TODO: take tap subnet parameter
        #TODO: this should be tuple of self.host and platform
        lab_topology = self.nidb.topology[self.host]
            # TODO: also store platform
        from netaddr import IPNetwork
        address_block = IPNetwork(settings.get("tapsn")
            or "172.16.0.0/16").iter_hosts() # added for backwards compatibility
        lab_topology.tap_host = address_block.next()
        lab_topology.tap_vm = address_block.next()  # for tunnel host
        for node in sorted(self.nidb.nodes("is_l3device", host=self.host)):
            # TODO: fix sorting order
            # TODO: check this works for switches
            node.tap.ip = address_block.next()

    def lab_topology(self):
# TODO: replace name/label and use attribute from subgraph
        lab_topology = self.nidb.topology[self.host]
        lab_topology.render_template = os.path.join("templates","netkit_lab_conf.mako")
        lab_topology.render_dst_folder = os.path.join("rendered", self.host, "netkit")
        lab_topology.render_dst_file = "lab.conf"
        lab_topology.description = "AutoNetkit Lab"
        lab_topology.author = "AutoNetkit"
        lab_topology.web = "www.autonetkit.org"
        host_nodes = list(
            self.nidb.nodes(host=self.host, platform="netkit"))
        if not len(host_nodes):
            log.debug("No Netkit hosts for %s" % self.host)
            # TODO: make so can return here
            # return
# also need collision domains for this host
        cd_nodes = self.nidb.nodes("collision_domain", host=self.host)
        host_nodes += cd_nodes
        subgraph = self.nidb.subgraph(host_nodes, self.host)

# TODO: sort this numerically, not just by string
        lab_topology.machines = " ".join(alpha_sort(naming.network_hostname(phy_node)
            for phy_node in subgraph.nodes("is_l3device")))

        lab_topology.config_items = []
        for node in sorted(subgraph.nodes("is_l3device")):
            for interface in node.physical_interfaces:
                collision_domain = str(interface.ipv4_subnet).replace("/", ".")
                numeric_id = interface.numeric_id #netkit lab.conf uses 1 instead of eth1
                lab_topology.config_items.append(
                    device=naming.network_hostname(node),
                    key=numeric_id,
                    value=collision_domain,
                )

        lab_topology.tap_ips = []
        for node in subgraph:
            if node.tap:
                lab_topology.tap_ips.append(
                    # TODO: merge the following and previous into a single
                    # function
                    device=naming.network_hostname(node),
                    id=node.tap.id.replace("eth", ""),  # strip ethx -> x
                    ip=node.tap.ip,
                )

        lab_topology.tap_ips.sort("ip")
        lab_topology.config_items.sort("device")


class CiscoCompiler(PlatformCompiler):
    """Platform compiler for Cisco"""

    # def __init__(self, nidb, anm, host):
# TODO: setup to remap allocate interface id function here
        # super(CiscoCompiler, self).__init__(nidb, anm, host)

    @staticmethod
    def loopback_interface_ids():
        for x in itertools.count(100):
            prefix = IosBaseCompiler.lo_interface_prefix
            yield "%s%s" % (prefix, x)

    @staticmethod
    def interface_ids_ios_by_slot():
        id_pairs = ((slot, 0) for slot in itertools.count(0))
        for (slot, port) in id_pairs:
            # yield "Ethernet%s/%s" % (slot, port)
            yield "GigabitEthernet%s/%s" % (slot, port)

    @staticmethod
    def interface_ids_ios():
        #TODO: make this skip if in list of allocated ie [interface.name for interface in node]
        for x in itertools.count(0):
            yield "GigabitEthernet0/%s" % x

    @staticmethod
    def interface_ids_ra():
        #TODO: make this skip if in list of allocated ie [interface.name for interface in node]
        for x in itertools.count(0):
            yield "GigabitEthernet%s" % x

    @staticmethod
    def interface_ids_nxos():
        for x in itertools.count(0):
            yield "Ethernet2/%s" % x

    @staticmethod
    def interface_ids_ios2_slot_port():
        """Allocate with slot and port iterating
        """
        id_pairs = ((slot, port) for (
            slot, port) in itertools.product(xrange(17), xrange(5)))
        for (slot, port) in id_pairs:
            yield "GigabitEthernet%s/%s/%s/%s" % (0, 0, slot, port)

    @staticmethod
    def interface_ids_ios2():
        for x in itertools.count(0):
            yield "GigabitEthernet0/0/0/%s" % x

    def compile(self):
        settings = autonetkit.config.settings
        to_memory = settings['Compiler']['Cisco']['to memory']
#TODO: need to copy across the interface name from edge to the interface
        g_phy = self.anm['phy']
        use_mgmt_interfaces = g_phy.data.mgmt_interfaces_enabled 
        if use_mgmt_interfaces:
            log.info("Allocating management interfaces for Cisco")
        else:
            log.info("Not allocating management interfaces for Cisco")

        log.info("Compiling Cisco for %s" % self.host)
        ios_compiler = IosClassicCompiler(self.nidb, self.anm)
        now = datetime.now()
        if settings['Compiler']['Cisco']['timestamp']:
            timestamp = now.strftime("%Y%m%d_%H%M%S_%f")
            dst_folder = os.path.join("rendered", self.host, timestamp, "cisco")
        else:
            dst_folder = os.path.join("rendered", self.host, "cisco")
# TODO: merge common router code, so end up with three loops: routers, ios
# routers, ios2 routers

        # store autonetkit_cisco version
        from pkg_resources import get_distribution
        ank_cisco_version = get_distribution("autonetkit_cisco").version

        for phy_node in g_phy.nodes('is_router', host=self.host):
            loopback_ids = self.loopback_interface_ids()
            # allocate loopbacks to routes (same for all ios variants)
            nidb_node = self.nidb.node(phy_node)
            nidb_node.ank_cisco_version = ank_cisco_version

            for interface in nidb_node.loopback_interfaces:
                if interface != nidb_node.loopback_zero:
                    interface.id = loopback_ids.next()

        for phy_node in g_phy.nodes('is_router', host=self.host, syntax='ios'):
            specified_int_names = phy_node.specified_int_names
            nidb_node = self.nidb.node(phy_node)
            nidb_node.render.template = os.path.join("templates","ios.mako")
            if to_memory:
                nidb_node.render.to_memory = True
            else:
                nidb_node.render.dst_folder = dst_folder
                nidb_node.render.dst_file = "%s.conf" % naming.network_hostname(
                    phy_node)

            # Assign interfaces
            if phy_node.device_subtype == "os":
                int_ids = self.interface_ids_ios()
            elif phy_node.device_subtype == "ra":
                int_ids = self.interface_ids_ra()
            else:
                # default if no subtype specified
                #TODO: need to set default in the load module
                log.warning("Unexpected subtype %s" % phy_node.device_subtype)
                int_ids = self.interface_ids_ios()
                
            if use_mgmt_interfaces:
                mgmt_int_id = int_ids.next()  # 0/0 is used for management ethernet

            for interface in nidb_node.physical_interfaces:
                if specified_int_names:
                    interface.id = phy_node.interface(interface).name
                # TODO: need to determine if interface name already specified
                else:
                    interface.id = int_ids.next()

            ios_compiler.compile(nidb_node)
            if use_mgmt_interfaces:
                mgmt_int = nidb_node.add_interface(management = True)
                mgmt_int.id = mgmt_int_id

        ios2_compiler = Ios2Compiler(self.nidb, self.anm)
        for phy_node in g_phy.nodes('is_router', host=self.host, syntax='ios2'):
            specified_int_names = phy_node.specified_int_names
            nidb_node = self.nidb.node(phy_node)
            nidb_node.render.template = os.path.join("templates","ios2","router.conf.mako")
            if to_memory:
                nidb_node.render.to_memory = True
            else:
                nidb_node.render.dst_folder = dst_folder
                nidb_node.render.dst_file = "%s.conf" % naming.network_hostname(
                    phy_node)

            # Assign interfaces
            int_ids = self.interface_ids_ios2()
            for interface in nidb_node.physical_interfaces:
                if specified_int_names:
                    interface.id = phy_node.interface(interface).name
                # TODO: need to determine if interface name already specified
                else:
                    interface.id = int_ids.next()

            ios2_compiler.compile(nidb_node)

            if use_mgmt_interfaces:
                mgmt_int_id = "mgmteth0/0/CPU0/0"
                mgmt_int = nidb_node.add_interface(management = True)
                mgmt_int.id = mgmt_int_id

        nxos_compiler = NxOsCompiler(self.nidb, self.anm)
        for phy_node in g_phy.nodes('is_router', host=self.host, syntax='nx_os'):
            nidb_node = self.nidb.node(phy_node)
            nidb_node.render.template = os.path.join("templates","nx_os.mako")
            if to_memory:
                nidb_node.render.to_memory = True
            else:
                nidb_node.render.dst_folder = dst_folder
                nidb_node.render.dst_file = "%s.conf" % naming.network_hostname(
                    phy_node)

            # Assign interfaces
            int_ids = self.interface_ids_nxos()
            for interface in nidb_node.physical_interfaces:
                if specified_int_names:
                    interface.id = phy_node.interface(interface).name
                # TODO: need to determine if interface name already specified
                else:
                    interface.id = int_ids.next()

            if use_mgmt_interfaces:
                mgmt_int_id = "mgmt0"
                mgmt_int = nidb_node.add_interface(management = True)
                mgmt_int.id = mgmt_int_id

            nxos_compiler.compile(nidb_node)

        other_nodes = [phy_node for phy_node in g_phy.nodes('is_router', host=self.host)
                       if phy_node.syntax not in ("ios", "ios2")]
        for node in other_nodes:
            #TODO: check why we need this
            phy_node = g_phy.node(node)
            nidb_node = self.nidb.node(phy_node)
            nidb_node.input_label = phy_node.id  # set specifically for now for other variants

# TODO: use more os.path.join for render folders
# TODO: Split compilers into seperate modules

        if use_mgmt_interfaces:
            self.assign_management_interfaces()

    def assign_management_interfaces(self):
        g_phy = self.anm['phy']
        lab_topology = self.nidb.topology[self.host]
        oob_management_ips = {}

        #TODO: make this seperate function
        from netaddr import IPNetwork, IPRange

        mgmt_address_start = g_phy.data.mgmt_address_start 
        mgmt_address_end = g_phy.data.mgmt_address_end 
        mgmt_prefixlen = int(g_phy.data.mgmt_prefixlen)

        #TODO: need to check if range is insufficient
        mgmt_ips = (IPRange(mgmt_address_start, mgmt_address_end))
        mgmt_ips_iter = iter(mgmt_ips) # to iterate over

        mgmt_address_start_network = IPNetwork(mgmt_address_start) # as /32 for supernet
        mgmt_address_end_network = IPNetwork(mgmt_address_end) # as /32 for supernet
        # retrieve the first supernet, as this is the range requested. subsequent are the subnets
        start_subnet = mgmt_address_start_network.supernet(mgmt_prefixlen)[0] # retrieve first
        end_subnet = mgmt_address_end_network.supernet(mgmt_prefixlen)[0] # retrieve first

        try: # validation
            assert(start_subnet == end_subnet)
            log.debug("Verified: Cisco management subnets match")
        except AssertionError:
            log.warning("Error: Cisco management subnets do not match: %s and %s, using start subnet"
                    % (start_subnet, end_subnet))

        mgmt_subnet = start_subnet
        hosts_to_allocate = sorted(self.nidb.nodes('is_router', host=self.host))
        dhcp_subtypes = {"os"}
        dhcp_hosts = [h for h in hosts_to_allocate if h.device_subtype in dhcp_subtypes]
        non_dhcp_hosts = [h for h in hosts_to_allocate if h.device_subtype not in dhcp_subtypes]

        try: # validation
            assert(len(mgmt_ips) >= len(non_dhcp_hosts))
            log.debug("Verified: Cisco management IP range is sufficient size %s for %s hosts"
                    % (len(mgmt_ips), len(non_dhcp_hosts)))
        except AssertionError:
            log.warning("Error: Cisco management IP range is insufficient size %s for %s hosts"
                    % (len(mgmt_ips), len(non_dhcp_hosts)))
            # TODO: need to use default range
            return

        for nidb_node in hosts_to_allocate:
            for interface in nidb_node.physical_interfaces:
                if interface.management:
                    interface.description = "OOB Management"
                    interface.physical = True
                    if nidb_node in dhcp_hosts:
                        interface.use_dhcp = True
                        oob_management_ips[str(nidb_node)] = "dhcp"
                    else:
                        ipv4_address = mgmt_ips_iter.next()
                        interface.ipv4_address = ipv4_address
                        interface.ipv4_subnet = mgmt_subnet
                        interface.ipv4_cidr = sn_preflen_to_network(ipv4_address, mgmt_prefixlen)
                        oob_management_ips[str(nidb_node)] = ipv4_address

        lab_topology.oob_management_ips = oob_management_ips


class DynagenCompiler(PlatformCompiler):
    """Dynagen Platform Compiler"""
    config_dir = "configs"

    @staticmethod
    def console_ports():
        """Interator for console ports"""
        for x in itertools.count(2001):
            yield x

    @staticmethod
    def interface_ids():
        """Allocate with slot and port iterating """
        id_pairs = ((slot + 1, port) for (
            slot, port) in itertools.product(xrange(4), xrange(2)))
        for (slot, port) in id_pairs:
            yield "f%s/%s" % (slot, port)

    def compile(self):
        log.info("Compiling Dynagen for %s" % self.host)
        g_phy = self.anm['phy']
        G_graphics = self.anm['graphics']
        ios_compiler = IosClassicCompiler(self.nidb, self.anm)
        for phy_node in g_phy.nodes('is_router', host=self.host, syntax='ios'):
            nidb_node = self.nidb.node(phy_node)
            graphics_node = G_graphics.node(phy_node)
            nidb_node.render.template = os.path.join("templates", "ios.mako")
            nidb_node.render.dst_folder = os.path.join(
                "rendered", self.host, "dynagen", self.config_dir)
            nidb_node.render.dst_file = "%s.cfg" % ank.name_folder_safe(
                phy_node.label)

            # TODO: may want to normalise x/y
            nidb_node.x = graphics_node.x
            nidb_node.y = graphics_node.y

            # Allocate edges
            # assign interfaces
            # Note this could take external data
            int_ids = self.interface_ids()
            for interface in nidb_node.physical_interfaces:
                interface.id = int_ids.next()

            ios_compiler.compile(nidb_node)

        self.allocate_ports()
        self.lab_topology()

    def allocate_ports(self):
        # TODO: take tap subnet parameter
        con_ports = self.console_ports()

        for node in sorted(self.nidb.nodes("is_l3device", host=self.host)):
            # TODO: fix sorting order
            # TODO: check this works for switches
            node.console_port = con_ports.next()
            node.aux_port = node.console_port + 500

    def lab_topology(self):
# TODO: replace name/label and use attribute from subgraph
        lab_topology = self.nidb.topology[self.host]
        lab_topology.render_template = os.path.join("templates","dynagen.mako")
        lab_topology.render_dst_folder = os.path.join("rendered", self.host, "dynagen")
        lab_topology.render_dst_file = "topology.net"

        lab_topology.config_dir = self.config_dir

        # TODO: pick these up from config
        lab_topology.hypervisor_server = "127.0.0.1"
        lab_topology.hypervisor_port = "7200"
        lab_topology.image = "router.image"
        lab_topology.idlepc = "0x60629004"

        lab_topology.routers = []
        routers = list(
            self.nidb.routers(host=self.host, platform="dynagen"))

        for router in routers:
            phy_node = self.anm['phy'].node(router)
            interfaces = []

            for link in phy_node.edges():
                nidb_edge = self.nidb.edge(link)
                # need to find the reverse link, and its place in nidb for port
# TODO: tidy this up once have interfaces implemented
                back_link = self.anm['phy'].edges(link.dst, phy_node).next()
                back_link_nidb = self.nidb.edge(back_link)
                interfaces.append({
                    'src_port': nidb_edge.id,
                    'dst': str(link.dst),
                    'dst_port': back_link_nidb.id,
                })

            slots = []
            import math
            number_of_slots = int(math.ceil(1.0 * len(interfaces) / 2))
            slots = [(index + 1, "PA-2FE-TX") for index in range(
                number_of_slots)]
            cnfg = os.path.join(self.config_dir, router.render.dst_file)

            lab_topology.routers.append(
                hostname=str(router),
                model=7200,
                console=router.console_port,
                aux=router.aux_port,
                interfaces=interfaces,
                x=router.x,
                y=router.y,
                slots=slots,
                cnfg=cnfg,
            )

        lab_topology.routers.sort("hostname")
        return
